/******************************************************************************

                               Copyright (c) 2007
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

#include "dsl_cpe_control.h"
#include "drv_dsl_cpe_api.h"
#include "drv_dsl_cpe_api_ioctl.h"
#ifdef INCLUDE_DSL_API_CLI_LEGACY

const char *sFailureReturn = "nReturn=%d";


/* for debugging: */
#ifdef DSL_CLI_LOCAL
#undef DSL_CLI_LOCAL
#endif
#if 0
#define DSL_CLI_LOCAL
#else
#define DSL_CLI_LOCAL static
#endif

DSL_Error_t
DSL_CPE_CLI_CheckHelp (const DSL_char_t * pCommands, const DSL_char_t * pUsage,
                   DSL_char_t * sOut, DSL_size_t nOutSz)
{
   if (pCommands
      && (strstr (pCommands, "-h") || strstr (pCommands, "--help")
      || strstr (pCommands, "/h") || strstr (pCommands, "-?")))
   {
      DSL_CPE_snprintf (sOut, nOutSz, pUsage);
      return DSL_ERROR;
   }
   return DSL_SUCCESS;
}

DSL_void_t
DSL_CPE_CLI_GetDeviceRange (DSL_uint16_t nDevice, DSL_uint16_t * nStart,
                        DSL_uint16_t * nEnd)
{
   if (nDevice == (DSL_uint16_t) - 1)
   {
      *nStart = 0;
      *nEnd = DSL_MAX_DEVICE_INSTANCES;
   }
   else
   {
      *nStart = nDevice;
      *nEnd = nDevice + 1;
   }
}

DSL_CLI_LOCAL DSL_void_t
DSL_CPE_CLI_G997_ChannelStatusGet (DSL_int32_t fd, DSL_char_t * pCommands,
                               DSL_char_t * sOut, DSL_size_t nOutSz)
{
   DSL_Error_t ret = 0;
   adslLineRateInfo info;
   DSL_uint16_t nLine=0;
   DSL_uint8_t nChannel=0;
   DSL_uint8_t nDirection=0;

#ifndef DSL_CPE_DEBUG_DISABLE
   static const DSL_char_t USAGE[] =
      "Long Form: G997_ChannelStatusGet" DSL_CPE_CRLF "Short Form: g997csg" DSL_CPE_CRLF
      "Output Parameter" DSL_CPE_CRLF
      "- DSL_Error_t nReturn" DSL_CPE_CRLF
      "- DSL_G997_ChannelStatus_t pData" DSL_CPE_CRLF DSL_CPE_CRLF "";
#else
#undef USAGE
#define USAGE ""
#endif /* DSL_CPE_DEBUG_DISABLE */

   if (DSL_CPE_CLI_CheckHelp (pCommands, USAGE, sOut, nOutSz) == DSL_ERROR)
   {
      return;
   }
   ret = DSL_CPE_sscanf (pCommands, "%bu %bu", &nChannel, &nDirection);
   if (ret != 2)
   {
      DSL_CPE_snprintf (sOut, nOutSz, USAGE);
      return;
   }

   memset(&info, 0x00, sizeof(info));

   info.adslDataRateus = 100000000;
   info.adslDataRateds = 100000000;

   info.flags = LINE_RATE_DATA_RATEDS_FLAG | LINE_RATE_DATA_RATEUS_FLAG;

   /*
   ret = DSL_CPE_Ioctl(fd, GET_ADSL_LINE_RATE, &info) ;
   if (ret < 0)
   {
   DSL_CPE_FPrintf(DSL_CPE_STDOUT, "ERROR: g997csg - can't execute GET_ADSL_LINE_RATE\n\r");
   goto EXECUTE_ERROR;
   }
   */
   if(nDirection == 0)
   {
      DSL_CPE_snprintf (sOut, nOutSz, "nReturn=0 nLine=%d nChannel=%d nDirection=%d ActualDataRate=%u "
         "PreviousDataRate=0 ActualInterleaveDelay=0 ActualImpulseNoiseProtection=0",
         nLine, nChannel, nDirection, info.adslDataRateus);
   }
   else if(nDirection == 1)
   {
      DSL_CPE_snprintf (sOut, nOutSz, "nReturn=0 nLine=%d nChannel=%d nDirection=%d ActualDataRate=%u "
         "PreviousDataRate=0 ActualInterleaveDelay=0 ActualImpulseNoiseProtection=0",
         nLine, nChannel, nDirection, info.adslDataRateds);
   }
}

DSL_CLI_LOCAL DSL_void_t
DSL_CPE_CLI_LineStateGet (DSL_int32_t fd, DSL_char_t * pCommands,
                      DSL_char_t * sOut, DSL_size_t nOutSz)
{
   DSL_Error_t ret = 0;
   DSL_uint16_t nLine=0;
   adslLineStatusInfo info;

#ifndef DSL_CPE_DEBUG_DISABLE
   static const DSL_char_t USAGE[] =
      "Long Form: LineStateGet" DSL_CPE_CRLF "Short Form: lsg" DSL_CPE_CRLF
      "Output Parameter" DSL_CPE_CRLF
      "- DSL_Error_t nReturn" DSL_CPE_CRLF
      "- DSL_LineState_t pData" DSL_CPE_CRLF DSL_CPE_CRLF "";
#else
#undef USAGE
#define USAGE ""
#endif /* DSL_CPE_DEBUG_DISABLE */

   if (DSL_CPE_CLI_CheckHelp (pCommands, USAGE, sOut, nOutSz) == DSL_ERROR)
   {
      return;
   }

   memset(&info, 0x00, sizeof(info));

#ifdef LINUX
   ret = DSL_CPE_Ioctl(fd, GET_ADSL_LINE_STATUS, &info) ;
   if (ret < 0)
   {
      return;
   }
#endif

   DSL_CPE_snprintf (sOut, nOutSz, "nReturn=0 nLine=%d nLineState=%d", nLine, 801);
}

DSL_CLI_LOCAL DSL_void_t
DSL_CPE_CLI_LineTrafficStatusGet (DSL_int32_t fd, DSL_char_t * pCommands,
                              DSL_char_t * sOut, DSL_size_t nOutSz)
{
   DSL_Error_t ret = 0;
   DSL_uint16_t nLine=0;
   DSL_uint8_t nChannel=0;
   DSL_uint8_t nDirection=0;
   static unsigned long nOctets=0;
   double tmp=0;

#ifndef DSL_CPE_DEBUG_DISABLE
   static const DSL_char_t USAGE[] =
      "Long Form: LineTrafficStatusGet" DSL_CPE_CRLF "Short Form: lsg" DSL_CPE_CRLF
      "Output Parameter" DSL_CPE_CRLF
      "- DSL_Error_t nReturn" DSL_CPE_CRLF
      "- DSL_LineState_t pData" DSL_CPE_CRLF DSL_CPE_CRLF "";
#else
#undef USAGE
#define USAGE ""
#endif /* DSL_CPE_DEBUG_DISABLE */

   if (DSL_CPE_CLI_CheckHelp (pCommands, USAGE, sOut, nOutSz) == DSL_ERROR)
   {
      return;
   }

   ret = DSL_CPE_sscanf (pCommands, "%bu %bu", &nChannel, &nDirection);
   if (ret != 2)
   {
      DSL_CPE_snprintf (sOut, nOutSz, USAGE);
      return;
   }

   tmp =  ((double)((unsigned long)rand()) / RAND_MAX) * 125000;
   nOctets += tmp;

   DSL_CPE_FPrintf(DSL_CPE_STDOUT, "nOctets=%u (%f)\n", nOctets, tmp);

   /*
   DSL_uint16_t msg[16];
   DSL_char_t group=0;
   DSL_uint16_t address=0;
   DSL_uint16_t index=0;
   DSL_uint32_t size=0;
   DSL_CmvRead(fd, group, address, index, size, msg, sizeof(msg));
   */
   DSL_CPE_snprintf (sOut, nOutSz, "nReturn=0 nLine=%d nDirection=%d nOctets=%u nLineRate=%u", nLine, nDirection, nOctets, 0);
}

DSL_CLI_LOCAL DSL_void_t
DSL_CPE_CLI_VersionInformationGet (DSL_int32_t fd, DSL_char_t * pCommands,
                               DSL_char_t * sOut, DSL_size_t nOutSz)
{
   DSL_Error_t ret = 0;
#ifndef DSL_CPE_DEBUG_DISABLE
   static const DSL_char_t USAGE[] =
      "Long Form: VersionInformationGet" DSL_CPE_CRLF "Short Form: vig" DSL_CPE_CRLF
      "Output Parameter" DSL_CPE_CRLF
      "- DSL_Error_t nReturn" DSL_CPE_CRLF
      "- DSL_VersionInformation_t pData" DSL_CPE_CRLF DSL_CPE_CRLF "";
#else
#undef USAGE
#define USAGE ""
#endif /* DSL_CPE_DEBUG_DISABLE */

   if (DSL_CPE_CLI_CheckHelp (pCommands, USAGE, sOut, nOutSz) == DSL_ERROR)
   {
      return;
   }

   DSL_CPE_snprintf (sOut, nOutSz, "nReturn=0 DSL_APILibraryVersion=1.9.5 DSL_ChipSetFWVersion=0.0.0.0.0.2 DSL_ChipSetHWVersion=Unknown DSL_ChipSetType=VDSL2_VINAX DSL_DriverVersion=0.0");
}

DSL_CliCommand_t cliCommand[] = {
   {"g997csg", "G997_ChannelStatusGet", DSL_CPE_CLI_G997_ChannelStatusGet},
   {"lsg", "LineStateGet", DSL_CPE_CLI_LineStateGet},
   {"ltsg", "LineTrafficStatusGet", DSL_CPE_CLI_LineTrafficStatusGet},
   {"vig", "VersionInformationGet", DSL_CPE_CLI_VersionInformationGet},
   {DSL_NULL, DSL_NULL, DSL_NULL}
};

#if 0
   else if(strcmp(cmd, "ltsg") == 0)
   {
   }
#endif

   static DSL_int32_t DSL_CmvMake(DSL_char_t opcode, DSL_char_t group, DSL_uint16_t address,
      DSL_uint16_t index, DSL_uint32_t size, DSL_uint16_t *pData, DSL_uint16_t *pMessage, DSL_uint32_t msg_len)
   {
      if ((msg_len < 16*2) || (pMessage==DSL_NULL))
         return -1;

      memset(pMessage, 0, 16*2);
      pMessage[0]= (opcode<<4) + (size&0xf);
      if(opcode == H2D_DEBUG_WRITE_DM)
         pMessage[1]= (group&0x7f);
      else
         pMessage[1]= (((index==0)?0:1)<<7) + (group&0x7f);
      pMessage[2]= address;
      pMessage[3]= index;
      if(pData)
      {
         if((opcode == H2D_CMV_WRITE)||(opcode == H2D_DEBUG_WRITE_DM))
            memcpy(pMessage+4, pData, size*2);
      }
      return 0;
   }

   static DSL_int32_t DSL_CmvRead(DSL_int32_t fd, DSL_char_t group, DSL_uint16_t address,
      DSL_uint16_t index, DSL_uint32_t size, DSL_uint16_t *pMessage, DSL_uint16_t msg_len)
   {
      if(DSL_CmvMake(H2D_CMV_READ, group, address, index, size, NULL, pMessage, msg_len) < 0)
      {
         return -1;
      }

#ifdef LINUX
      if(DSL_CPE_Ioctl(fd, DANUBE_MEI_CMV_WINHOST, &pMessage) < 0)
      {
         return -1;
      }
#endif

      return 0;
   }

#endif /* INCLUDE_DSL_API_CLI */
