/******************************************************************************
**
** FILE NAME    : dwc_otg_ifx.h
** PROJECT      : Twinpass/Danube
** MODULES      : DWC OTG USB
**
** DATE         : 12 April 2007
** AUTHOR       : Sung Winder
** DESCRIPTION  : Platform specific initialization.
** COPYRIGHT    :       Copyright (c) 2007
**                      Infineon Technologies AG
**                      2F, No.2, Li-Hsin Rd., Hsinchu Science Park,
**                      Hsin-chu City, 300 Taiwan.
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date          $Author         $Comment
** 12 April 2007   Sung Winder     Initiate Version
*******************************************************************************/
#if !defined(__DWC_OTG_IFX_H__)
#define __DWC_OTG_IFX_H__

// 20070316, winder added.
#if defined(CONFIG_DANUBE)
#include <asm/danube/danube.h>

/**
 * This function is called to set correct clock gating and power.
 * For Twinpass/Danube board.
 */
#define DANUBE_RCU_UBSCFG  ((volatile u32*)(DANUBE_RCU_BASE_ADDR + 0x18))
#define DANUBE_RCU_RESET   ((volatile u32*)(DANUBE_RCU_BASE_ADDR + 0x10))
#define DANUBE_USBCFG_HDSEL_BIT    11	// 0:host, 1:device
#define DANUBE_USBCFG_HOST_END_BIT 10	// 0:little_end, 1:big_end

#define IFX_USB_IRQ 54

static void dwc_otg_power_on (void)
{
	// clear power
	set_bit (0, DANUBE_PMU_PWDCR);
	set_bit (6, DANUBE_PMU_PWDCR);

	MDELAY(100);

    // set clock gating
    set_bit (4, DANUBE_CGU_IFCCR);
    set_bit (5, DANUBE_CGU_IFCCR);

	// set power
	clear_bit (0, DANUBE_PMU_PWDCR);
	clear_bit (6, DANUBE_PMU_PWDCR);
    clear_bit (15, DANUBE_PMU_PWDCR);

    set_bit (4, DANUBE_RCU_RESET);
	MDELAY(100);
    clear_bit (4, DANUBE_RCU_RESET);

#if defined (DWC_HOST_ONLY)
    // make the hardware be a host controller (default)
    clear_bit (DANUBE_USBCFG_HDSEL_BIT, DANUBE_RCU_UBSCFG);
#elif defined (DWC_DEVICE_ONLY)
    /* set the controller to the device mode */
    set_bit (DANUBE_USBCFG_HDSEL_BIT, DANUBE_RCU_UBSCFG);
#else
#error  "For Danube/Twinpass, it should be HOST or Device Only."
#endif

    // set the HC's byte-order to big-endian
    set_bit (DANUBE_USBCFG_HOST_END_BIT, DANUBE_RCU_UBSCFG);

    // PHY configurations.
    writel (0x14014, 0xbe10103c);
}


static void dwc_otg_phy_power_on (void)
{
	// set power
	clear_bit (0, DANUBE_PMU_PWDCR);
    writel (0x14014, 0xbe10103c);
}

static void dwc_otg_phy_power_off (void)
{
	// set power
	set_bit (0, DANUBE_PMU_PWDCR);
    writel (0x14014, 0xbe10103c);
}


#elif defined(CONFIG_AMAZON_SE) // CONFIG_DANUBE
#include <asm/amazon_se/amazon_se.h>
#define IFX_USB_IRQ 31


/**
 * This function is called to set correct clock gating and power.
 * For Amazon-SE board.
 */
#define AMAZON_SE_RCU_UBSCFG  ((volatile u32*)(AMAZON_SE_RCU_BASE_ADDR + 0x18))
#define AMAZON_SE_RCU_RESET   ((volatile u32*)(AMAZON_SE_RCU_BASE_ADDR + 0x10))
#define AMAZON_SE_USBCFG_HDSEL_BIT    11	// 0:host, 1:device
#define AMAZON_SE_USBCFG_HOST_END_BIT 10	// 0:little_end, 1:big_end
static void dwc_otg_power_on (void)
{
	// clear power
	set_bit (0, AMAZON_SE_PMU_PWDCR);
	set_bit (6, AMAZON_SE_PMU_PWDCR);

	MDELAY(100);

    // set power
    clear_bit (0, AMAZON_SE_PMU_PWDCR);
    clear_bit (6, AMAZON_SE_PMU_PWDCR);


    clear_bit (15, AMAZON_SE_PMU_PWDCR);

    set_bit (4, AMAZON_SE_RCU_RESET);
	MDELAY(500);
    clear_bit (4, AMAZON_SE_RCU_RESET);
	MDELAY(500);

#if defined (DWC_HOST_ONLY)
    // make the hardware be a host controller (default)
    clear_bit (AMAZON_SE_USBCFG_HDSEL_BIT, AMAZON_SE_RCU_UBSCFG);
#elif defined (DWC_DEVICE_ONLY)
    /* set the controller to the device mode */
    set_bit (AMAZON_SE_USBCFG_HDSEL_BIT, AMAZON_SE_RCU_UBSCFG);
#else
#error  "For Danube/Twinpass, it should be HOST or Device Only."
#endif

    // set the HC's byte-order to big-endian
    set_bit (AMAZON_SE_USBCFG_HOST_END_BIT, AMAZON_SE_RCU_UBSCFG);


    // PHY configurations.
    writel (0x14014, 0xbe10103c);
}
static void dwc_otg_phy_power_on (void)
{
    clear_bit (0, AMAZON_SE_PMU_PWDCR);
    writel (0x14014, 0xbe10103c);
}

static void dwc_otg_phy_power_off (void)
{
    set_bit (0, AMAZON_SE_PMU_PWDCR);
    writel (0x14014, 0xbe10103c);
}


#else // CONFIG_AMAZON_SE
#error "Please choose one platform!!"
#endif // CONFIG_AMAZON_SE

#endif //__DWC_OTG_IFX_H__
