/* This is a module which is used for setting the NFMARK field of an skb. */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/checksum.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_SCVLANMARK.h>

static unsigned int
target(struct sk_buff **pskb,
       unsigned int hooknum,
       const struct net_device *in,
       const struct net_device *out,
       const void *targinfo,
       void *userinfo)
{
	const struct ipt_scvlanmark_target_info *markinfo = targinfo;

	(*pskb)->sc_out_8021p = markinfo->mark;
	(*pskb)->sc_out_8021p_enable = markinfo->enable;

	return IPT_CONTINUE;
}

static int
checkentry(const char *tablename,
	   const struct ipt_entry *e,
           void *targinfo,
           unsigned int targinfosize,
           unsigned int hook_mask)
{
	if (targinfosize != IPT_ALIGN(sizeof(struct ipt_scvlanmark_target_info))) {
		printk(KERN_WARNING "MARK: targinfosize %u != %Zu\n",
		       targinfosize,
		       IPT_ALIGN(sizeof(struct ipt_scvlanmark_target_info)));
		return 0;
	}



	return 1;
}

static struct ipt_target ipt_scvlanmark_reg
= { { NULL, NULL }, "SCVLANMARK", target, checkentry, NULL, THIS_MODULE };

static int __init init(void)
{
	if (ipt_register_target(&ipt_scvlanmark_reg))
		return -EINVAL;

	printk("scvlanmark init\n");

	return 0;
}

static void __exit fini(void)
{
	ipt_unregister_target(&ipt_scvlanmark_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");

