/******************************************************************************

                               Copyright (c) 2007
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

#define DSL_INTERN

#include "drv_dsl_cpe_api.h"
#include "drv_dsl_cpe_api_ioctl.h"

#ifdef __cplusplus
   extern "C" {
#endif

#undef DSL_DBG_BLOCK
#define DSL_DBG_BLOCK DSL_DBG_CPE_API

/* handles to the low-level BSP driver */
DSL_devCtx_t ifxDevices[DSL_MAX_DEVICE_INSTANCES] = { {DSL_NULL} };

#ifdef DSL_CPE_STATIC_DELT_DATA
   DSL_G997_DeltData_t deltData;
#ifdef INCLUDE_DSL_CPE_API_VINAX
   DSL_G997_DeltShowtimeData_t deltShowtimeData;
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#endif /* DSL_CPE_STATIC_DELT_DATA*/

DSL_OpenContext_t *pOpenContextList = DSL_NULL;
extern DSL_uint32_t DSL_DRV_LED_ModuleInit(DSL_Context_t *pContext);

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_EventGenerate(
   DSL_Context_t *pContext,
   DSL_uint8_t nChannel,
   DSL_AccessDir_t nAccessDir,
   DSL_XTUDir_t nXtuDir,
   DSL_EventType_t nEventType,
   DSL_EventData_Union_t *pData,
   DSL_uint16_t nDataSize)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_EventStatusData_t evData;

   /* Generate event*/
   memset(&evData, 0, sizeof(DSL_EventStatusData_t));
   evData.nEventType = nEventType;
   evData.nXtuDir    = nXtuDir;
   evData.nAccessDir = nAccessDir;
   evData.nChannel   = nChannel;

   if( pData != DSL_NULL && nDataSize != 0)
   {
      evData.pData = (DSL_EventData_Union_t*)DSL_DRV_Malloc(nDataSize);
   }
   else
   {
      evData.pData = DSL_NULL;
   }

   if( evData.pData != DSL_NULL)
   {
      memcpy( evData.pData, pData, nDataSize);
   }

   /* Queue Event*/
   nErrCode = DSL_DRV_EventQueue(pContext, &evData, nDataSize);

   return nErrCode;
}

/*
   This function hadles an LINIT event.

   \param pContext - Pointer to dsl cpe library context structure, [I]
   \param nLinit   - Specifies the LINIT failure definition, [I]

   \return
   Return values are defined within the DSL_Error_t definition
   - DSL_SUCCESS in case of success
   - DSL_ERROR if operation failed
*/
DSL_Error_t DSL_DRV_HandleLinitValue(
   DSL_Context_t *pContext,
   const DSL_G997_LineInit_t nLinit,
   const DSL_G997_LineInitSubStatus_t nSub)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_G997_LineInitStatusData_t lineInitStatus;
   DSL_uint32_t nCount = 0;

   lineInitStatus.nLineInitStatus    = nLinit;
   lineInitStatus.nLineInitSubStatus = nSub;

   DSL_CTX_WRITE(pContext, nErrCode, lineInitStatus, lineInitStatus);

   if (nLinit != LINIT_SUCCESSFUL)
   {
      /* Get LINIT Retry count*/
      DSL_CTX_READ_SCALAR(pContext, nErrCode, lineInitRetryCount, nCount);

      if( nCount < DSL_G997_LINIT_RETRIES )
      {
         nCount++;
      }

      if( (nCount >= DSL_G997_LINIT_RETRIES) || (nLinit == LINIT_NO_PEER_XTU) )
      {
         nErrCode = DSL_DRV_EventGenerate(
                       pContext, 0, DSL_ACCESSDIR_NA, DSL_XTUDIR_NA,
                       DSL_EVENT_S_LINIT_FAILURE,
                       (DSL_EventData_Union_t*)&nLinit,
                       sizeof(DSL_G997_LineInit_t));

         if( nErrCode != DSL_SUCCESS )
         {
            DSL_DEBUG( DSL_DBG_ERR,
               (pContext, "DSL[%02d]: ERROR - Event(%d) generate failed!"DSL_DRV_CRLF,
               DSL_DEV_NUM(pContext), DSL_EVENT_S_LINIT_FAILURE));
         }
      }
   }
   else
   {
      /* Reset LINIT retry counter in case of success */
      nCount = 0;
   }

   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, lineInitRetryCount, nCount);

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_OnTimeoutEvent(
   DSL_Context_t *pContext,
   DSL_int_t nEventType,
   DSL_uint32_t nTimeoutID)
{
   DSL_Error_t errorCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,(pContext, "DSL[%02d]: DSL_DRV_OnTimeoutEvent: "
      "nEventType=%d, nTimeoutID=0x%08X" DSL_DRV_CRLF ,
      DSL_DEV_NUM(pContext), nEventType, nTimeoutID));

   errorCode = DSL_DRV_DEV_OnTimeoutEvent(pContext,nEventType,nTimeoutID);

   return (errorCode);
}

static DSL_Error_t DSL_DRV_FwCopy(
   DSL_Context_t *pContext,
   DSL_boolean_t bIsInKernel,
   const DSL_uint8_t *pFirmware,
   const DSL_uint32_t nFirmwareSize,
   DSL_uint8_t **ppFirmware,
   DSL_uint32_t *pFirmwareSize)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, ppFirmware);
   DSL_CHECK_POINTER(pContext, pFirmwareSize);
   DSL_CHECK_ERR_CODE();

   if ( (pFirmware != DSL_NULL) && (nFirmwareSize != 0) )
   {
      if (*ppFirmware != DSL_NULL)
      {
         DSL_DRV_VFree(*ppFirmware);
      }

      *ppFirmware = (DSL_uint8_t*)DSL_DRV_VMalloc(nFirmwareSize);
      if (*ppFirmware != DSL_NULL)
      {
         *pFirmwareSize = nFirmwareSize;
         /* Copy FW binary*/
         DSL_IoctlMemCpyFrom(bIsInKernel, *ppFirmware, (DSL_void_t*)pFirmware,
            nFirmwareSize);
      }
      else
      {
         *pFirmwareSize = 0;
         nErrCode = DSL_ERROR;
         DSL_DEBUG( DSL_DBG_ERR, (pContext,
            "DSL[%02d]: Memory allocation for firmware binary failed!" DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
      }
   }

   return (nErrCode);
}

/*
   This function stores the FW binaries in the DSL CPE API internal memory

   \param pContext      Pointer to dsl library context structure, [I]
   \param pFirmware1    1st firmware pointer, [I]
   \param nSize1        Size of 1st firmware, [I]
   \param pFirmware2    2nd firmware pointer, [I]
   \param nSize2        Size of 2nd firmware, [I]

   \return  Return values are defined within the \ref DSL_Error_t definition
    - DSL_SUCCESS in case of success
    - DSL_ERROR if operation failed
    - or any other defined specific error code
*/
static DSL_Error_t DSL_DRV_FwStore(
   DSL_Context_t *pContext,
   DSL_boolean_t bIsInKernel,
   const DSL_char_t *pFirmware1,
   DSL_uint32_t nSize1,
   const DSL_char_t *pFirmware2,
   DSL_uint32_t nSize2)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_FwStore"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Proceed 1'st FW binary*/
   if (DSL_DRV_FwCopy(pContext, bIsInKernel, pFirmware1, nSize1,
      &pContext->pFirmware, &pContext->nFirmwareSize) != DSL_SUCCESS)
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_API_VINAX
   /* Proceed 2'nd FW binary*/
   if( DSL_DRV_FwCopy(pContext, bIsInKernel, pFirmware2, nSize2,
          &pContext->pFirmware2, &pContext->nFirmwareSize2) != DSL_SUCCESS)
   {
      return DSL_ERROR;
   }
#endif /* #ifdef INCLUDE_DSL_CPE_API_VINAX*/


   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_FwStore"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return DSL_SUCCESS;
}

static DSL_Error_t DSL_DRV_InstanceStatusGet(
   DSL_OpenContext_t *pOpenCtx,
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_InstanceStatus_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_InstanceStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pOpenCtx->eventMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]:Couldn't lock event mutex"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   /* Get status data from the instance specific context*/   
   pData->data.bEventActivation        = pOpenCtx->bEventActivation;
   pData->data.nResourceActivationMask = pOpenCtx->nResourceActivationMask;

   DSL_DRV_MUTEX_UNLOCK(pOpenCtx->eventMutex);
   
   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_InstanceStatusGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return(nErrCode);
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_HandleInit(
   DSL_IN     DSL_int_t nNum,
   DSL_IN_OUT DSL_OpenContext_t **pRefOpenContext)
{
   if (nNum >= DSL_MAX_DEVICE_INSTANCES || pRefOpenContext == DSL_NULL)
   {
      return DSL_ERR_INVALID_PARAMETR;
   }

   /* the global context is NULL, initialize it */
   if (ifxDevices[nNum].pContext == DSL_NULL)
   {
      ifxDevices[nNum].pContext = DSL_DRV_VMalloc(sizeof(DSL_Context_t));
      if (ifxDevices[nNum].pContext == DSL_NULL)
      {
         return DSL_ERR_MEMORY;
      }

      DSL_DRV_MemSet(ifxDevices[nNum].pContext, 0, sizeof(DSL_Context_t));

      if (ifxDevices[nNum].nUsageCount != 0)
      {
         DSL_DEBUG(DSL_DBG_ERR, (DSL_NULL, "DSL[%02d]: Internal error in the DSL "
                            "CPE API. Uninitialized device context has non-null "
                            "usage count."DSL_DRV_CRLF, nNum));
         return DSL_ERR_INTERNAL;
      }

      ifxDevices[nNum].pContext->pDevCtx = &ifxDevices[nNum];

      /* Set device number*/
      ifxDevices[nNum].nNum = nNum;
      /* init device mutex */
      DSL_DRV_MUTEX_INIT(ifxDevices[nNum].pContext->bspMutex);

      /* init data mutex */
      DSL_DRV_MUTEX_INIT(ifxDevices[nNum].pContext->dataMutex);

      /* init initialization mutex */
      DSL_DRV_MUTEX_INIT(ifxDevices[nNum].pContext->initMutex);

      if (DSL_DRV_Timeout_Init(ifxDevices[nNum].pContext, DSL_MAX_TIMEOUT_NUM,
         DSL_FALSE) != DSL_SUCCESS)
      {
         DSL_DEBUG(DSL_DBG_ERR, (DSL_NULL, "DSL[%02d]: Internal error in the DSL "
                            "CPE API. Could not prepare timeout context."DSL_DRV_CRLF,
                            nNum));
         return DSL_ERR_INTERNAL;
      }
   }

   if (ifxDevices[nNum].nUsageCount > 0 && ifxDevices[nNum].pContext == DSL_NULL)
   {
      DSL_DEBUG(DSL_DBG_ERR, (DSL_NULL, "DSL[%02d]: Internal error in the DSL "
                         "CPE API. Used device context has null pointer "
                         "to DSL Context."DSL_DRV_CRLF,
                         nNum));
      return DSL_ERR_INTERNAL;
   }

   *pRefOpenContext = DSL_DRV_VMalloc(sizeof(DSL_OpenContext_t));
   if (*pRefOpenContext == DSL_NULL)
   {
      DSL_DEBUG(DSL_DBG_ERR, (DSL_NULL, "DSL[%02d]: Open failed, "
         "could not allocate memory..."DSL_DRV_CRLF, nNum));
      return DSL_ERR_MEMORY;
   }
   else
   {
      DSL_DRV_MemSet(*pRefOpenContext, 0, sizeof(DSL_OpenContext_t));

      /* init event queue */
      DSL_DRV_INIT_WAKELIST("evntque", (*pRefOpenContext)->eventWaitQueue);

      /* init event mutex */
      DSL_DRV_MUTEX_INIT((*pRefOpenContext)->eventMutex);

      #ifdef INCLUDE_DSL_CEOC
      /* init Rx SNMP FIFO mutex*/
      DSL_DRV_MUTEX_INIT((*pRefOpenContext)->rxSnmpFifoMutex);
      #endif /* #ifdef INCLUDE_DSL_CEOC*/

      /* Disable event handling*/
      (*pRefOpenContext)->bEventActivation = DSL_FALSE;
      /* Mask (disable) all events*/
      (*pRefOpenContext)->nEventMask       = 0xFFFFFFFF;
      /* KAv: Unmask all indication threshold crossing events by default.
         It not completely defined how to handle these events for instance specific
         processing.*/
      (*pRefOpenContext)->nEventMask &= 
         ~(DSL_EVENT2MASK(DSL_EVENT_I_LINE_THRESHOLD_CROSSING) |
           DSL_EVENT2MASK(DSL_EVENT_I_CHANNEL_THRESHOLD_CROSSING) |
           DSL_EVENT2MASK(DSL_EVENT_I_DATA_PATH_THRESHOLD_CROSSING) |
           DSL_EVENT2MASK(DSL_EVENT_I_CHANNEL_DATARATE_SHIFT_THRESHOLD_CROSSING));

      (*pRefOpenContext)->nResourceActivationMask = 0xFFFFFFFF;

      /* add to list */
      if (pOpenContextList == DSL_NULL)
      {
         /* first element */
         pOpenContextList = *pRefOpenContext;
      }
      else
      {
         /* add to head */
         (*pRefOpenContext)->pNext = pOpenContextList;
         pOpenContextList = *pRefOpenContext;
      }

      (*pRefOpenContext)->pDevCtx = &ifxDevices[nNum];

      ifxDevices[nNum].nUsageCount++;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (DSL_NULL, "DSL[%02d]: Open: Use count: %d"DSL_DRV_CRLF,
      nNum, ifxDevices[nNum].nUsageCount));

   DSL_DRV_OS_ModUseCountIncrement();

   return DSL_SUCCESS;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_HandleCleanup(
   DSL_IN DSL_devCtx_t *pDevContext,
   DSL_IN DSL_boolean_t bForce)
{
   if (pDevContext == DSL_NULL)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (DSL_NULL, "DSL[%02d]: DSL_DRV_HandleCleanup: "
         "Device context is NULL"DSL_DRV_CRLF, pDevContext->nNum));
      return DSL_ERROR;
   }
   
   if (pDevContext->nUsageCount > 1 && bForce == DSL_FALSE)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (DSL_NULL, "DSL[%02d]: DSL_DRV_HandleCleanup: "
         "device is still used by another process"DSL_DRV_CRLF,
         pDevContext->nNum));
      return DSL_ERROR;
   }

   if (pDevContext->nUsageCount > 1 && bForce == DSL_TRUE)
   {
      DSL_DEBUG( DSL_DBG_MSG,
         (DSL_NULL, "DSL[%02d]: DSL_DRV_HandleCleanup: "
         "clean up is forced"DSL_DRV_CRLF, pDevContext->nNum));
   }

   if (pDevContext->nUsageCount > 1 || bForce == DSL_TRUE)
   {

      while (pDevContext->nUsageCount-- > 1)
      {
         DSL_DRV_OS_ModUseCountDecrement();
      }

      if (pDevContext->pContext != DSL_NULL)
      {
         DSL_DEBUG( DSL_DBG_MSG,
            (DSL_NULL, "DSL[%02d]: DSL_DRV_HandleCleanup: "
            "Autoboot thread stopping..."DSL_DRV_CRLF,
            pDevContext->nNum));

         DSL_DRV_AutobootThreadStop(pDevContext->pContext);

         DSL_DEBUG( DSL_DBG_MSG,
            (DSL_NULL, "DSL[%02d]: DSL_DRV_HandleCleanup: "
            "Autoboot thread has been stopped..."DSL_DRV_CRLF,
            pDevContext->nNum));

         DSL_DRV_Free(pDevContext->pContext);
         DSL_DRV_VFree(pDevContext->pContext);
         pDevContext->pContext = DSL_NULL;
      }
   }

   return DSL_SUCCESS;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_HandleDelete(
   DSL_IN DSL_OpenContext_t *pOpenContext)
{
   DSL_devCtx_t *pDevContext;
   DSL_EventStatusData_t event;
   DSL_OpenContext_t *pPrev, *pCurr;
   DSL_void_t *pPtr;
   DSL_Error_t nErrCode;

   if (pOpenContext == DSL_NULL)
   {
      return DSL_ERR_INVALID_PARAMETR;
   }
   else
   {
      pDevContext = pOpenContext->pDevCtx;

      if (pDevContext == DSL_NULL)
      {
         return DSL_ERROR;
      }
      else
      {
         if (pDevContext->nUsageCount > 0)
         {
            pDevContext->nUsageCount--;
         }

         if ((pOpenContext->nEventMask & DSL_EVENT2MASK(DSL_EVENT_S_FIRMWARE_REQUEST)) == 0
            && pOpenContext->pDevCtx != DSL_NULL
            && pOpenContext->pDevCtx->pContext != DSL_NULL)
         {
            DSL_CTX_WRITE_SCALAR(pOpenContext->pDevCtx->pContext,
               nErrCode, bFirmwareEventAssigned, DSL_FALSE);
         }

         if (pOpenContext->eventFifo != DSL_NULL
            && pOpenContext->eventFifoBuf != DSL_NULL)
         {
            while (!DSL_Fifo_isEmpty(pOpenContext->eventFifo))
            {
               if (DSL_DRV_EventUnqueue(pOpenContext, &event)
                  == DSL_SUCCESS)
               {
                  if (event.pData != DSL_NULL)
                  {
                     DSL_DRV_VFree(event.pData);
                  }
               }
            }
         }

         /* release event structures */
         if (pOpenContext->eventFifo != DSL_NULL)
         {
            pPtr = pOpenContext->eventFifo;
            pOpenContext->eventFifo = DSL_NULL;

            DSL_DRV_VFree(pPtr);
         }

         /* release event structures */
         if (pOpenContext->eventFifoBuf != DSL_NULL)
         {
            pPtr = pOpenContext->eventFifoBuf;
            pOpenContext->eventFifoBuf = DSL_NULL;
            DSL_DRV_VFree(pPtr);
         }

         /* remove from list */
         if (pOpenContextList != DSL_NULL)
         {
            if (pOpenContext == pOpenContextList)
            {
               /* it is a head */
               pOpenContextList = pOpenContext->pNext;
            }
            else
            {
               for ( pPrev = pOpenContextList,
                     pCurr = pOpenContextList->pNext;
                     pCurr != pOpenContext && pCurr != DSL_NULL;
                     pPrev = pCurr,
                     pCurr = pCurr->pNext);

               if (pCurr != DSL_NULL)
               {
                  pPrev->pNext = pCurr->pNext;
               }
            }
         }

         DSL_DRV_VFree(pOpenContext);
      }
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (DSL_NULL, "DSL[%02d]: Close: Use count: %d"DSL_DRV_CRLF,
      pDevContext->nNum, pDevContext->nUsageCount));

   DSL_DRV_OS_ModUseCountDecrement();

   return DSL_SUCCESS;
}

DSL_void_t DSL_DRV_Cleanup(DSL_void_t)
{
   DSL_int_t i;

   for (i = 0; i < DSL_MAX_DEVICE_INSTANCES; i++)
   {
      DSL_DRV_HandleCleanup(&ifxDevices[i], DSL_TRUE);

      DSL_DRV_DEV_DriverHandleDelete(ifxDevices[i].lowHandle);

#ifdef INCLUDE_DSL_CPE_API_VINAX
      DSL_DRV_DEV_DriverHandleDelete(ifxDevices[i].nfc_lowHandle);
#endif

      /* this is redundant code but in case of some our
         threads leave active it will prevent momentary fault */
      ifxDevices[i].lowHandle = DSL_NULL;
#ifdef INCLUDE_DSL_CPE_API_VINAX
      ifxDevices[i].nfc_lowHandle = DSL_NULL;
#endif
      ifxDevices[i].nUsageCount = 0;
   }

   while (pOpenContextList != DSL_NULL)
   {
      DSL_DRV_HandleDelete(pOpenContextList);
   }
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_LinkReset(
   DSL_IN DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_LinkReset"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   nErrCode = DSL_DRV_DEV_LinkReset(pContext);
   if (nErrCode != DSL_SUCCESS)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: Modem couldn't be reset"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_MSG,
         (pContext, "DSL[%02d]: Modem was reseted..."DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_SUCCESS;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_LinkReset"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

DSL_Error_t DSL_DRV_ModulesInit(
   DSL_IN DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_boolean_t bFirmwareReady = DSL_FALSE;

   DSL_CHECK_CTX_POINTER(pContext);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_ModulesInit"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CTX_READ_SCALAR(pContext, nErrCode, pDevCtx->bFirmwareReady, bFirmwareReady);

   /* Check FW availability*/
   if( !bFirmwareReady )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - FW not loaded yet!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Lock DSL CPE API data mutex*/
   if(DSL_DRV_MUTEX_LOCK(pContext->initMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: Couldn't lock data mutex"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

#if defined(INCLUDE_ADSL_LED) && (defined(INCLUDE_DSL_CPE_API_AMAZON_SE) \
                               || defined(INCLUDE_DSL_CPE_API_AMAZON_S))
   if ( DSL_DRV_LED_ModuleInit(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - LED module initialization failed"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      nErrCode = DSL_ERROR;
   }
#endif

#if defined(INCLUDE_DSL_ADSL_MIB)
   if ( DSL_DRV_MIB_ModuleInit(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - MIB module initialization failed"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }
#endif

#if defined(INCLUDE_DSL_CEOC)
   if ( DSL_CEOC_Start(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - CEOC module initialization failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }
#endif

#if defined(INCLUDE_DSL_PM)
   if ( DSL_DRV_PM_Start(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM module initialization failed"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }
#endif

   /* Unock DSL CPE API data mutex*/
   DSL_DRV_MUTEX_UNLOCK(pContext->initMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_ModulesInit"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_XtseSettingsCheck(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN DSL_uint8_t *pXTSE)
{
   DSL_Error_t nErrCode = DSL_SUCCESS, nRet = DSL_SUCCESS;
   DSL_int32_t i = 0, emptyXseCnt = 0;
   DSL_uint8_t xtsePoss = 0;
   DSL_boolean_t bXtseChanged = DSL_FALSE, bXtseUnsupported = DSL_FALSE;

   DSL_CHECK_POINTER(pContext, pXTSE);
   DSL_CHECK_ERR_CODE();
   
   /* Validation checks for configuration data */
   for (i = 0; i < DSL_G997_NUM_XTSE_OCTETS; i++)
   {
      DSL_CTX_READ(pContext, nErrCode, xtsePoss[i], xtsePoss);

      /* Check for unsupported bits*/
      if ( (~(xtsePoss) & pXTSE[i]) > 0)
      {
         /* Check if there is at least one valid bit*/
         if ((pXTSE[i] & xtsePoss) > 0)
         {
            /* Change specified configuration*/
            pXTSE[i] &= xtsePoss;
            
            bXtseChanged = DSL_TRUE;
         }
         else
         { 
            /* No valid bits found at all*/
            bXtseUnsupported = DSL_TRUE;
         }
      }

      if (!pXTSE[i])
      {
         emptyXseCnt++;
      }
   }

   /* Check if no valid bits specified at all*/
   if (emptyXseCnt >= DSL_G997_NUM_XTSE_OCTETS)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Empty XTSE settings, emptyXseCnt=%d!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), emptyXseCnt));

      nErrCode = DSL_ERROR;
   }
   else
   {
      if (bXtseUnsupported && !bXtseChanged)
      {
         /* There are bits set within the configuration that are not supported
            by G.997.1 or at least by this version of the DSL library */
         DSL_DEBUG(DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - No valid XTSE settings!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));

         nErrCode = DSL_ERROR;
      }
      else if (bXtseChanged)
      {
         DSL_DEBUG(DSL_DBG_WRN,
            (pContext, "DSL[%02d]: WARNING - Unsupported bits removed from the current XTSE settings!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
         
         nErrCode = DSL_WRN_INCONSISTENT_XTSE_CONFIGURATION;
      }
   }

   if (nErrCode >= DSL_SUCCESS)
   {
      /* Check device specific XTSE settings*/
      nRet = DSL_DRV_DEV_XtseSettingsCheck(pContext, pXTSE);
      if (nRet != DSL_SUCCESS)
      {
         nErrCode = nRet;
      }
   }   

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_Init(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_Init_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int32_t nLoff = 0, nCurrOffset = 0;
   DSL_boolean_t bFirmwareReady = DSL_FALSE;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_Init"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_POINTER(pContext, pContext->pDevCtx);
   DSL_CHECK_ERR_CODE();

   if (pContext->bInitComplete == DSL_TRUE)
   {
      DSL_DEBUG( DSL_DBG_WRN,
         (pContext, "DSL[%02d]: DSL CPE API is already initialized!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_WRN_ALREADY_INITIALIZED;
   }

   memcpy( pContext->xtseCfg, pData->data.nXtseCfg.XTSE, sizeof(pContext->xtseCfg) );

   /* Initialize possible XTSE configurations*/
#if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[1-1], 0x17);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[2-1], 0x01);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[3-1], 0x14);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[4-1], 0x51);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[5-1], 0x4C);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[6-1], 0x45);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[7-1], 0x05);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[8-1], 0x07);
#else
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[1-1], 0x17);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[2-1], 0x01);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[3-1], 0x14);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[4-1], 0x51);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[5-1], 0x4C);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[6-1], 0x45);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[7-1], 0x05);
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, xtsePoss[8-1], 0x00);
#endif

   /* Get Test Mode Control options and copy them to the DSL CPE context*/
   memcpy(&(pContext->nTestModeControl),
          &(pData->data.nTestModeControl), sizeof(DSL_TestModeControlSet_t));

   /* Get Line Activate Control options and copy them to the DSL CPE context*/
   memcpy(&(pContext->lineActivateConfig),
          &(pData->data.nLineActivateConfig), sizeof(DSL_G997_LineActivateData_t));


#ifdef DSL_CPE_STATIC_DELT_DATA
   memset(&deltData, 0, sizeof(DSL_G997_DeltData_t));
   pContext->DELT = &deltData;
#ifdef INCLUDE_DSL_CPE_API_VINAX
   memset(&deltShowtimeData, 0, sizeof(DSL_G997_DeltShowtimeData_t));
   pContext->DELT_SHOWTIME = &deltShowtimeData;
#endif   
#else
   pContext->DELT = DSL_NULL;
#ifdef INCLUDE_DSL_CPE_API_VINAX
   pContext->DELT_SHOWTIME = DSL_NULL;
#endif   
#endif

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: Couldn't lock data mutex"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   /* Set SerialNumber*/
   memcpy(
      &pContext->lineInventoryNe.SerialNumber,
      pData->data.pInventory->SerialNumber,
      DSL_G997_LI_MAXLEN_SERIAL);

   /* Set SystemVendorID*/
   memcpy(
      &pContext->lineInventoryNe.SystemVendorID,
      pData->data.pInventory->SystemVendorID,
      DSL_G997_LI_MAXLEN_VENDOR_ID);

   /* Set VersionNumber*/
   memcpy(
      &pContext->lineInventoryNe.VersionNumber,
      pData->data.pInventory->VersionNumber,
      DSL_G997_LI_MAXLEN_VERSION);
#if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)
   /* Set VersionNumber*/
   memcpy(
      &pContext->auxInventoryNe,
      &(pData->data.pInventory->Auxiliary),
      sizeof(DSL_AuxInventoryNe_t));
#endif

   /* Set Autoboot configuration options*/
   pContext->nAutobootStartupMode = pData->data.nAutobootStartupMode;
   memcpy(
      &(pContext->nAutobootConfig),
      &(pData->data.nAutobootConfig),
      sizeof(DSL_AutobootConfigData_t));

   /* TODO: check these initialization parameters*/
   pContext->lineFeatureDataCfg[DSL_UPSTREAM].bTrellisEnable = DSL_TRUE;
   pContext->lineFeatureDataCfg[DSL_DOWNSTREAM].bTrellisEnable = DSL_TRUE;
   pContext->lineFeatureDataCfg[DSL_UPSTREAM].bBitswapEnable = DSL_TRUE;
   pContext->lineFeatureDataCfg[DSL_DOWNSTREAM].bBitswapEnable = DSL_TRUE;
   pContext->powerMgmtStatus.nPowerManagementStatus = DSL_G997_PMS_L3;

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   /* Check XTSE setings*/
   nErrCode =  DSL_DRV_XtseSettingsCheck(pContext, pData->data.nXtseCfg.XTSE);
   if (nErrCode < DSL_SUCCESS)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Wrong XTSE settings!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      memset( pContext->xtseCfg, 0x0, sizeof(pContext->xtseCfg) );

      return nErrCode;
   }

   /* Initialize device specific parameters*/
   nErrCode = DSL_DRV_DeviceInit(pContext, pData);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Device Init failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      return nErrCode;
   }

   /* Download FW if available and Start Autoboot handling.
       If no FW specified, do it later with appropriate IOCTLs*/
   if( pData->data.pFirmware != DSL_NULL || pData->data.pFirmware2 != DSL_NULL )
   {
      /* Load FW if available*/
      nErrCode = DSL_DRV_FwDownload(
                    pContext,
                    pData->data.pFirmware,
                    pData->data.nFirmwareSize,
                    pData->data.pFirmware2,
                    pData->data.nFirmwareSize2,
                    &nLoff, &nCurrOffset,
                    DSL_TRUE);

      if( nErrCode != DSL_SUCCESS )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - FW Download failed!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
         return nErrCode;
      }

      /* Get FW ready flag*/
      DSL_CTX_READ_SCALAR(pContext, nErrCode, pDevCtx->bFirmwareReady, bFirmwareReady);

      /* Here the complete FW binary should be downloaded to the device, no chunks allowed*/
      if(!bFirmwareReady)
	   {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - Expecting complete FW binary!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         return DSL_ERROR;
      }

      /* Start Autoboot handling if allowed*/
      if( pData->data.nAutobootStartupMode != DSL_AUTOBOOT_CTRL_STOP )
      {
         if( DSL_DRV_AutobootThreadStart(pContext) != DSL_SUCCESS )
         {
            DSL_DEBUG( DSL_DBG_ERR,
               (pContext, "DSL[%02d]: ERROR - Autoboot thread start failed!"DSL_DRV_CRLF,
               DSL_DEV_NUM(pContext)));
               
            return DSL_ERROR;
         }
      }

      /* Initialize additional modules PM, CEOC,...*/
      nErrCode = DSL_DRV_ModulesInit(pContext);
      if( nErrCode != DSL_SUCCESS )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - additional modules init failed!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         return nErrCode;
      }
   }

   /* Set LINIT value to LINIT_UNKNOWN*/
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, lineInitStatus.nLineInitStatus, LINIT_UNKNOWN);

   pContext->bInitComplete = DSL_TRUE;

   /* Generate DSL_EVENT_S_INIT_READY event*/
   DSL_DRV_EventGenerate(pContext, 0, DSL_ACCESSDIR_NA, DSL_XTUDIR_NA,
      DSL_EVENT_S_INIT_READY, DSL_NULL, 0);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_Init, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext) ,nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_DeviceInit(
   DSL_Context_t *pContext,
   DSL_Init_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: IN - DSL_DRV_DeviceInit"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   nErrCode = DSL_DRV_DEV_DeviceInit(pContext, pData);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_DeviceInit, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_void_t DSL_DRV_Free(
   DSL_IN DSL_Context_t *pContext)
{
   if (pContext == DSL_NULL)
   {
      DSL_DEBUG_HDR(DSL_DBG_ERR, (pContext, "Invalid context pointer!"));
      return;
   }

   if (pContext->pFirmware != DSL_NULL)
   {
      DSL_DRV_VFree(pContext->pFirmware);
      pContext->pFirmware = DSL_NULL;
   }

   if (pContext->pFirmware2 != DSL_NULL)
   {
      DSL_DRV_VFree(pContext->pFirmware2);
      pContext->pFirmware2 = DSL_NULL;
   }

   pContext->bInitComplete = DSL_FALSE;

   DSL_DRV_Timeout_Shutdown(pContext);

#ifdef INCLUDE_DSL_PM
   /* Stop PM module processing*/
   if( DSL_DRV_PM_Stop(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: PM Module deinitialization failed"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
#endif /* #ifdef INCLUDE_DSL_PM*/

#ifdef INCLUDE_DSL_CEOC
   /* Stop CEOC module processing*/
   if( DSL_CEOC_Stop(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: CEOC Module deinitialization failed"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
#endif /* #ifdef INCLUDE_DSL_CEOC*/

#ifndef DSL_CPE_STATIC_DELT_DATA
   /* Release DELT memory */
   if (pContext->DELT != DSL_NULL)
   {
      DSL_DRV_MemFree(pContext->DELT);
   }
#ifdef INCLUDE_DSL_CPE_API_VINAX
   if (pContext->DELT_SHOWTIME != DSL_NULL)
   {
      DSL_DRV_MemFree(pContext->DELT_SHOWTIME);
   }
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#endif /* DSL_CPE_STATIC_DELT_DATA*/
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_AutobootLoadFirmware(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_AutobootLoadFirmware_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int32_t nOffset = 0;
   DSL_boolean_t bFirmwareReady = DSL_FALSE;
   DSL_boolean_t bAutobootThreadStarted = DSL_FALSE,
                 bAutobootFwLoadPending = DSL_FALSE;
   DSL_AutobootControl_t AutobootCtrl;
   DSL_int32_t nWaitCount = 0;
   DSL_Autoboot_State_t nState = DSL_AUTOBOOTSTATE_UNKNOWN;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_POINTER(pContext, pContext->pDevCtx);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ_SCALAR(pContext, nErrCode, bAutobootThreadStarted, bAutobootThreadStarted);
   DSL_CTX_READ_SCALAR(pContext, nErrCode, bAutobootFwLoadPending, bAutobootFwLoadPending);

   if( (pData->data.pFirmware  != DSL_NULL && pData->data.nFirmwareSize) ||
       (pData->data.pFirmware2 != DSL_NULL && pData->data.nFirmwareSize2) )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: Preparing to download FW..."DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
   else
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - No FW specified!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }


   if( bAutobootThreadStarted && !bAutobootFwLoadPending)
   {
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootFwLoadPending, DSL_TRUE);

      /* Get Autoboot State*/
      DSL_CTX_READ_SCALAR(pContext, nErrCode, nAutobootState, nState);

      if( nState != DSL_AUTOBOOTSTATE_FIRMWARE_WAIT )
      {
         /* Restart Autoboot*/
         AutobootCtrl.data.nCommand = DSL_AUTOBOOT_CTRL_RESTART;
         nErrCode = DSL_DRV_AutobootControlSet(pContext,&AutobootCtrl);
         if( nErrCode != DSL_SUCCESS )
         {
            /* Clear bAutobootFwLoadPending flag*/
            DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootFwLoadPending, DSL_FALSE);

            DSL_DEBUG(DSL_DBG_ERR,
               (pContext, "DSL[%02d]: ERROR - Autoboot Control set failed!"DSL_DRV_CRLF,
               DSL_DEV_NUM(pContext)));
            return nErrCode;
         }
      }

      /* wait for "DSL_AUTOBOOTSTATE_FIRMWARE_WAIT" max 4 sec */
      for (nWaitCount = 0; nWaitCount < 40; nWaitCount++)
      {
         /* Get Autoboot State*/
         DSL_CTX_READ_SCALAR(pContext, nErrCode, nAutobootState, nState);
         if( nErrCode != DSL_SUCCESS )
         {
            /* Clear bAutobootFwLoadPending flag*/
            DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootFwLoadPending, DSL_FALSE);

            DSL_DEBUG(DSL_DBG_ERR,
               (pContext, "DSL[%02d]: ERROR - Autoboot State get failed!"DSL_DRV_CRLF,
               DSL_DEV_NUM(pContext)));
            return nErrCode;
         }

         if( nState == DSL_AUTOBOOTSTATE_FIRMWARE_WAIT )
         {
            break;
         }

         DSL_DRV_MSecSleep(100);
      }

      if( nState != DSL_AUTOBOOTSTATE_FIRMWARE_WAIT )
      {
         /* Clear bAutobootFwLoadPending flag*/
         DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootFwLoadPending, DSL_FALSE);
         DSL_DEBUG(DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - Timeout while waiting the "
            "DSL_AUTOBOOTSTATE_FIRMWARE_WAIT state!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
         return DSL_ERROR;
      }
   }

   /* Download FW*/
   nErrCode = DSL_DRV_FwDownload(pContext, pData->data.pFirmware,
      pData->data.nFirmwareSize, pData->data.pFirmware2,
      pData->data.nFirmwareSize2, &pData->data.nFirmwareOffset, &nOffset,
      pData->data.bLastChunk);

   if( nErrCode == DSL_SUCCESS )
   {
      /* Get bFirmwareReady flag*/
      DSL_CTX_READ_SCALAR(pContext, nErrCode, pDevCtx->bFirmwareReady, bFirmwareReady);
      if( bFirmwareReady )
      {
         DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootFwLoadPending, DSL_FALSE);
         DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bFwRequestHandled, DSL_TRUE);
      }
   }
   else
   {
      DSL_DEBUG(DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - FW download failed, "
            "Autoboot will stay in the DSL_AUTOBOOTSTATE_FIRMWARE_WAIT state!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
   }

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_AutobootControlSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_AutobootControl_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_G997_LDSF_t nLoopMode = 0;
   DSL_boolean_t bFirmwareReady, bInitComplete;
   DSL_Autoboot_State_t nState = DSL_AUTOBOOTSTATE_UNKNOWN;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_AutobootControlSet(%d)"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), pData->data.nCommand));

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ_SCALAR(pContext, nErrCode, bInitComplete, bInitComplete);
   if (nErrCode == DSL_SUCCESS && bInitComplete == DSL_FALSE)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - DSL CPE API was not initialized yet!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      nErrCode = DSL_ERROR;
   }

   if (nErrCode == DSL_SUCCESS)
   {
      switch (pData->data.nCommand)
      {
         case DSL_AUTOBOOT_CTRL_START:
            DSL_CTX_READ_SCALAR(pContext, nErrCode, pDevCtx->bFirmwareReady,
               bFirmwareReady);
            if (bFirmwareReady == DSL_FALSE)
            {
               DSL_DEBUG(DSL_DBG_ERR,
                  (pContext, "DSL[%02d]: ERROR - Autoboot Start failed, no FW loaded!"DSL_DRV_CRLF,
                  DSL_DEV_NUM(pContext)));
               nErrCode = DSL_ERROR;
            }

            if( nErrCode == DSL_SUCCESS )
            {
               nErrCode = DSL_DRV_AutobootThreadStart(pContext);
               if( nErrCode != DSL_SUCCESS )
               {
                  DSL_DEBUG(DSL_DBG_ERR,
                    (pContext, "DSL[%02d]: ERROR - Autoboot Thread start failed!"DSL_DRV_CRLF,
                    DSL_DEV_NUM(pContext)));
                  break;
               }
            }

            /* Initialize additional modules PM, CEOC,...*/
            if( nErrCode == DSL_SUCCESS )
            {
               nErrCode = DSL_DRV_ModulesInit(pContext);
               if( nErrCode != DSL_SUCCESS )
               {
                  DSL_DEBUG( DSL_DBG_ERR,
                     (pContext, "DSL[%02d]: ERROR - additional modules init failed!"DSL_DRV_CRLF,
                     DSL_DEV_NUM(pContext)));
                  break;
               }
            }
            break;

         case DSL_AUTOBOOT_CTRL_STOP:
            /* Stop Autoboot thread*/
            DSL_DRV_AutobootThreadStop(pContext);
#ifdef INCLUDE_DSL_PM
            /* Suspend PM module*/
            nErrCode = DSL_DRV_PM_Suspend(pContext);

            if (nErrCode != DSL_SUCCESS)
            {
               DSL_DEBUG( DSL_DBG_ERR,
                  (pContext, "DSL[%02d]: ERROR - PM module suspend failed!"DSL_DRV_CRLF,
                  DSL_DEV_NUM(pContext)));
               
               return nErrCode;
            }
#endif /* #ifdef INCLUDE_DSL_PM*/
            /* Trigger restart sequence*/
            DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootRestart, DSL_TRUE);
            break;
            
         case DSL_AUTOBOOT_CTRL_RESTART:
            DSL_CTX_READ_SCALAR(pContext, nErrCode, lineActivateConfig.nLDSF, nLoopMode);

            if (nLoopMode == DSL_G997_AUTO_LDSF)
            {
               DSL_CTX_WRITE_SCALAR(pContext, nErrCode, nLoopAutoCount, 1);
            }

            DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootRestart, DSL_TRUE);

            if (pContext->bAutobootThreadStarted)
            {
               /* Change Autoboot Status*/
               nErrCode = DSL_DRV_AutobootStatusSet(pContext,
                             DSL_AUTOBOOT_STATUS_RUNNING, DSL_FW_REQUEST_NA);
            }

            break;

         case DSL_AUTOBOOT_CTRL_CONTINUE:
            DSL_CTX_READ_SCALAR(pContext, nErrCode, nAutobootState, nState);
            if ((nState != DSL_AUTOBOOTSTATE_LINK_ACTIVATE_WAIT) &&
                (nState != DSL_AUTOBOOTSTATE_CONFIG_WRITE_WAIT) &&
                (nState != DSL_AUTOBOOTSTATE_RESTART_WAIT))
            {
               nErrCode = DSL_WRN_NOT_ALLOWED_IN_CURRENT_STATE;
            }
            else
            {
               /* Trigger to continue Autoboot handling*/
               DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootContinue, DSL_TRUE);         
            }
            break;
         default:
            nErrCode = DSL_ERR_INVALID_PARAMETR;
            break;
      }
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_AutobootControlSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_AutobootConfigGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_OUT DSL_AutobootConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_AutobootConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Get Autoboot Configuration from the DSL CPE context*/
   DSL_CTX_READ(pContext, nErrCode, nAutobootConfig, pData->data);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_AutobootConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_AutobootConfigSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_OUT DSL_AutobootConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_AutobootConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Set Autoboot Configuration from the DSL CPE context*/
   DSL_CTX_WRITE(pContext, nErrCode, nAutobootConfig, pData->data);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_AutobootConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_AutobootStatusGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_OUT DSL_AutobootStatus_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_AutobootStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Get Autoboot Status from the DSL CPE context*/
   DSL_CTX_READ(pContext, nErrCode, nAutobootStatus.nStatus, pData->data.nStatus);

   DSL_CTX_READ(pContext, nErrCode, nAutobootStatus.nFirmwareRequestType,
      pData->data.nFirmwareRequestType);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_AutobootStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_AutobootStatusSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN DSL_AutobootStatGet_t nStatus,
   DSL_IN DSL_FirmwareRequestType_t nFirmwareRequestType)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_AutobootStatGet_t nOldStatus;
   DSL_AutobootStatusData_t autobootStatus;

   DSL_CHECK_CTX_POINTER(pContext);;
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_AutobootStatusSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CTX_READ_SCALAR(pContext, nErrCode, nAutobootStatus.nStatus, nOldStatus);

   if (nStatus != nOldStatus)
   {
      DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: Autoboot status has changed: "
         "%s(%d) -> %s(%d)" DSL_DRV_CRLF, DSL_DEV_NUM(pContext),
         DSL_DBG_PRN_AUTOBOOT_STATUS[nOldStatus], nOldStatus,
         DSL_DBG_PRN_AUTOBOOT_STATUS[nStatus], nStatus));

      autobootStatus.nStatus = nStatus;
      autobootStatus.nFirmwareRequestType = nFirmwareRequestType;
      
      /* Set Autoboot Status in the DSL CPE context*/
      DSL_CTX_WRITE(pContext, nErrCode, nAutobootStatus, autobootStatus);

      nErrCode =  DSL_DRV_EventGenerate(
         pContext, 0, DSL_ACCESSDIR_NA, DSL_XTUDIR_NA,
         DSL_EVENT_S_AUTOBOOT_STATUS,
         (DSL_EventData_Union_t*)&autobootStatus,
         sizeof(DSL_AutobootStatusData_t));

      if( nErrCode != DSL_SUCCESS )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - Event(%d) generate failed!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext), DSL_EVENT_S_AUTOBOOT_STATUS));
      }
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_AutobootStatusSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_VersionInformationGet(
    DSL_IN DSL_Context_t *pContext,
    DSL_IN_OUT DSL_VersionInformation_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_boolean_t bFirmwareReady = DSL_FALSE;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DRV_MemSet(pData, 0, sizeof(DSL_VersionInformation_t));
   strncpy(pData->data.DSL_DriverVersionApi, DSL_CPE_API_PACKAGE_VERSION,
           MAX_INFO_STRING_LEN);

   nErrCode = DSL_DRV_DEV_DriverVersionGet(pContext, pData->data.DSL_DriverVersionMeiBsp);
   if (nErrCode == DSL_SUCCESS)
   {
      strncpy(pData->data.DSL_ChipSetType, DSL_CHIPSET_TYPE_STRING, MAX_INFO_STRING_LEN);
   }

   if (nErrCode >= DSL_SUCCESS)
   {
      DSL_CTX_READ_SCALAR(pContext, nErrCode, pDevCtx->bFirmwareReady, bFirmwareReady);

      if (bFirmwareReady != DSL_FALSE)
      {
         nErrCode = DSL_DRV_DEV_FirmwareVersionGet(pContext,
            pData->data.DSL_ChipSetFWVersion);

         if (nErrCode >= DSL_SUCCESS)
         {
            nErrCode = DSL_DRV_DEV_ChipHardwareVersionGet(pContext,
               pData->data.DSL_ChipSetHWVersion);
         }
      }
      else
      {
         strcpy(pData->data.DSL_ChipSetFWVersion, "n/a");
         strcpy(pData->data.DSL_ChipSetHWVersion, "n/a");
      }
   }

   return (nErrCode);
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_LineStateGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_LineState_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LineStateGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, pData->data.nLineState);

   DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: OUT - DSL_DRV_LineStateGet,"
      " nLineState=%08X, retCode=%d"DSL_DRV_CRLF, DSL_DEV_NUM(pContext),
      pData->data.nLineState,nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_LineFeatureConfigSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_LineFeature_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LineFeatureConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_DIRECTION(pData->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_WRITE(pContext, nErrCode, lineFeatureDataCfg[pData->nDirection], pData->data);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LineFeatureConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_LineFeatureConfigGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_LineFeature_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LineFeatureConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_DIRECTION(pData->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ(pContext, nErrCode, lineFeatureDataCfg[pData->nDirection], pData->data);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LineFeatureConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_LineFeatureStatusGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_LineFeature_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LineFeatureStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_DIRECTION(pData->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ(pContext, nErrCode, lineFeatureDataSts[pData->nDirection], pData->data);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LineFeatureStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_InteropFeatureConfigSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_InteropFeatureConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_InteropFeatureConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

#ifdef INCLUDE_DSL_CPE_API_VINAX
   nErrCode = DSL_ERR_NOT_SUPPORTED_BY_FIRMWARE;
#else
   DSL_CTX_WRITE(pContext, nErrCode, interopFeatureConfigData, pData->data);
#endif /* INCLUDE_DSL_CPE_API_VINAX*/

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_InteropFeatureConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_InteropFeatureConfigGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_InteropFeatureConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_InteropFeatureConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

#ifdef INCLUDE_DSL_CPE_API_VINAX
   nErrCode = DSL_ERR_NOT_SUPPORTED_BY_FIRMWARE;
#else
   DSL_CTX_READ(pContext, nErrCode, interopFeatureConfigData, pData->data);
#endif /* INCLUDE_DSL_CPE_API_VINAX*/

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_InteropFeatureConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SYSTEM_INTERFACE_CONFIG_SET
*/
DSL_uint32_t DSL_DRV_SystemInterfaceConfigSet(
   DSL_Context_t *pContext,
   DSL_SystemInterfaceConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_SystemInterfaceConfigSet(nSystemIf=%#x, nTcLayer=%#x)" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), pData->data.nSystemIf, pData->data.nTcLayer));

   if ( ((pData->data.nSystemIf > DSL_SYSTEMIF_UNKNOWN) && 
      (pData->data.nSystemIf < DSL_SYSTEMIF_LAST)) ||
      ((pData->data.nTcLayer > DSL_TC_UNKNOWN) &&
      (pData->data.nTcLayer < DSL_TC_LAST)) )
   {
      nErrCode = DSL_DRV_DEV_SystemInterfaceConfigSet(pContext, pData);
   }
   else
   {
      nErrCode = DSL_ERR_PARAM_RANGE;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_SystemInterfaceConfigSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SYSTEM_INTERFACE_CONFIG_GET
*/
DSL_uint32_t DSL_DRV_SystemInterfaceConfigGet(
   DSL_Context_t *pContext,
   DSL_SystemInterfaceConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SystemInterfaceConfigGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   nErrCode = DSL_DRV_DEV_SystemInterfaceConfigGet(pContext, pData);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SystemInterfaceConfigGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_SYSTEM_INTERFACE_STATUS_GET
*/
DSL_uint32_t DSL_DRV_SystemInterfaceStatusGet(
   DSL_Context_t *pContext,
   DSL_SystemInterfaceStatus_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_SystemInterfaceStatusGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   nErrCode = DSL_DRV_DEV_SystemInterfaceStatusGet(pContext, pData);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_SystemInterfaceStatusGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

#ifdef INCLUDE_DSL_CPE_API_VINAX
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_LinePathCounterTotalGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_LinePathCounterTotal_t *pData )
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LinePathCounterTotalGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_LinePathCounterTotalGet(
                pContext, pData->nChannel, &(pData->data));
   
   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LinePathCounterTotalGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
   
   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_DataPathCounterTotalGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_DataPathCounterTotal_t *pData )
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_CHANNEL_RANGE(pData->nChannel);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_DataPathCounterTotalGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_DataPathCounterTotalGet(
                pContext, pData->nChannel, &(pData->data));
   
   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_DataPathCounterTotalGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
   
   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_LOW_LEVEL_CONFIGURATION_SET
*/
DSL_uint32_t DSL_DRV_LowLevelConfigurationSet(
   DSL_Context_t *pContext,
   DSL_LowLevelConfiguration_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LowLevelConfigurationSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_LowLevelConfigurationSet(pContext, &(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LowLevelConfigurationSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_LOW_LEVEL_CONFIGURATION_GET
*/
DSL_uint32_t DSL_DRV_LowLevelConfigurationGet(
   DSL_Context_t *pContext,
   DSL_LowLevelConfiguration_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LowLevelConfigurationGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Get Low Level Configuration from the VINAX device context*/
   DSL_CTX_READ(pContext, nErrCode, pDevCtx->data.deviceCfg.cfg, pData->data);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LowLevelConfigurationGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_api.h'
*/
DSL_uint32_t DSL_DRV_AuxLineInventoryGet(
   DSL_Context_t *pContext,
   DSL_AuxLineInventory_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_DIRECTION(pData->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_AuxLineInventoryGet(nDirection=%s)" DSL_DRV_CRLF ,
      DSL_DEV_NUM(pContext), (pData->nDirection==DSL_NEAR_END?"NE":"FE")));

   if (DSL_DRV_VNX_FwFeatureCheck(pContext, DSL_VNX_FW_VDSL2))
   {
      if (pData->nDirection == DSL_NEAR_END)
      {
         DSL_CTX_READ(pContext, nErrCode, auxInventoryNe, pData->data);
      }
      else
      {
         DSL_CTX_READ(pContext, nErrCode, auxInventoryFe, pData->data);
      }
   }
   else
   {
      nErrCode = DSL_ERR_NOT_SUPPORTED_BY_FIRMWARE;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_AuxLineInventoryGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_BAND_PLAN_STATUS_GET
*/
DSL_uint32_t DSL_DRV_BandPlanStatusGet(
   DSL_Context_t *pContext,
   DSL_BandPlanStatus_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_BandPlanStatusGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call VINAX device specific implementation*/
   nErrCode = DSL_DRV_VNX_BandPlanStatusGet(pContext,&(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_BandPlanStatusGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_BAND_PLAN_SUPPORT_GET
*/
DSL_uint32_t DSL_DRV_BandPlanSupportedGet(
   DSL_Context_t *pContext,
   DSL_BandPlanSupport_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_BandPlanSupportedGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call VINAX device specific implementation*/
   nErrCode = DSL_DRV_VNX_BandPlanSupportedGet(pContext, &(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_BandPlanSupportedGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_EFM_MAC_CONFIG_SET
*/
DSL_uint32_t DSL_DRV_EfmMacConfigSet(
   DSL_Context_t *pContext,
   DSL_EFM_MacConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_DIRECTION(pData->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_EfmMacConfigSet(nDirection=%d)" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), pData->nDirection));

   /* Check nEfmSpeed parameter validity*/
   if (pData->data.nEfmSpeed < DSL_EFM_SPEED_100 ||
       pData->data.nEfmSpeed >= DSL_EFM_SPEED_LAST)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Wrong EfmSpeed=%d specified" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pData->data.nEfmSpeed));
      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }

   /* Check nEfmDuplex parameter validity*/
   if (pData->data.nEfmDuplex < DSL_EFM_DUPLEX_FULL ||
       pData->data.nEfmDuplex >= DSL_EFM_DUPLEX_LAST)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Wrong nEfmDuplex=%d specified" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pData->data.nEfmDuplex));
      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }

   /* Check nFlowControl parameter validity*/
   if (pData->data.nFlowControl < DSL_EFM_FLOWCTRL_OFF ||
       pData->data.nFlowControl >= DSL_EFM_FLOWCTRL_LAST)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Wrong nFlowControl=%d specified" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pData->data.nFlowControl));
      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }

   /* Check nAutoNegotiation parameter validity*/
   if (pData->data.nAutoNegotiation < DSL_EFM_AUTONEG_OFF ||
       pData->data.nAutoNegotiation >= DSL_EFM_AUTONEG_LAST)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Wrong nAutoNegotiation=%d specified" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pData->data.nAutoNegotiation));
      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }

   /* Copy EFM MAC Configuration data to internal VINAX device context*/
   if (nErrCode == DSL_SUCCESS)
   {
      DSL_CTX_WRITE(pContext, nErrCode, pDevCtx->data.deviceCfg.MacCfg, pData->data);
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_EfmMacConfigSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_EFM_MAC_CONFIG_GET
*/
DSL_uint32_t DSL_DRV_EfmMacConfigGet(
   DSL_Context_t *pContext,
   DSL_EFM_MacConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_DIRECTION(pData->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_EfmMacConfigGet(nDirection=%d)" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), pData->nDirection));

   /* Get EFM MAC Configuration data from the internal VINAX device context*/
   DSL_CTX_READ(pContext, nErrCode, pDevCtx->data.deviceCfg.MacCfg, pData->data);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_EfmMacConfigGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_UTOPIA_BUS_WIDTH_CONFIG_SET
*/
DSL_uint32_t DSL_DRV_UtopiaBusWidthConfigSet(
   DSL_Context_t *pContext,
   DSL_UtopiaBusWidthConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_UtopiaBusWidthConfigSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_UtopiaBusWidthConfigSet(pContext, &(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_UtopiaBusWidthConfigSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_UTOPIA_BUS_WIDTH_CONFIG_GET
*/
DSL_uint32_t DSL_DRV_UtopiaBusWidthConfigGet(
   DSL_Context_t *pContext,
   DSL_UtopiaBusWidthConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_UtopiaBusWidthConfigGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_UtopiaBusWidthConfigGet(pContext, &(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_UtopiaBusWidthConfigGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}
#endif /* #ifdef INCLUDE_DSL_CPE_API_VINAX*/

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_UTOPIA_ADDRESS_CONFIG_SET
*/
DSL_uint32_t DSL_DRV_UtopiaAddressConfigSet(
   DSL_Context_t *pContext,
   DSL_PhyAddressConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_UtopiaAddressConfigSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_UtopiaAddressConfigSet(pContext,pData->nChannel,&(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_UtopiaAddressConfigSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_UTOPIA_ADDRESS_CONFIG_GET
*/
DSL_uint32_t DSL_DRV_UtopiaAddressConfigGet(
   DSL_Context_t *pContext,
   DSL_PhyAddressConfig_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_UtopiaAddressConfigGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific implementation*/
   nErrCode = DSL_DRV_DEV_UtopiaAddressConfigGet(pContext,pData->nChannel,&(pData->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_UtopiaAddressConfigGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return (nErrCode);
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_EVENT_STATUS_MASK_CONFIG_SET
*/
DSL_uint32_t DSL_DRV_EventStatusMaskConfigSet(
   DSL_OpenContext_t *pOpenCtx,
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_EventStatusMask_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint32_t nEventMask;
   DSL_boolean_t bFwEv = DSL_FALSE;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_EventStatusMaskConfigSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pOpenCtx->eventMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: Couldn't lock event mutex"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   /* Check if the event handling is activated for this instance*/
   if (pOpenCtx->bEventActivation)
   {
      /* Get current event mask*/
      nEventMask = pOpenCtx->nEventMask;

      switch(pData->data.nEventType)
      {
      case DSL_EVENT_S_FIRMWARE_REQUEST:
         DSL_CTX_READ_SCALAR(pContext, nErrCode, bFirmwareEventAssigned, bFwEv);
      
         if (pData->data.bMask)
         {
            /* Mask (disable) DSL_EVENT_S_FIRMWARE_REQUEST event*/
            nEventMask |= DSL_EVENT2MASK(DSL_EVENT_S_FIRMWARE_REQUEST);
            if (bFwEv != DSL_FALSE)
            {
               DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bFirmwareEventAssigned, DSL_FALSE);         
            }
         }
         else
         {
            if (bFwEv != DSL_FALSE)
            {
               DSL_DEBUG(DSL_DBG_WRN,
                  (DSL_NULL, "DSL[%02d]: Could not assign more than 1 applications "
                             "which handle firmware request event! Event mask will "
                             "be reduced"DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));
               /* Mask (disable) DSL_EVENT_S_FIRMWARE_REQUEST event*/
               nEventMask |= DSL_EVENT2MASK(DSL_EVENT_S_FIRMWARE_REQUEST);
            }
            else
            {
               /* Unmask (enable) DSL_EVENT_S_FIRMWARE_REQUEST event*/
               nEventMask &= ~DSL_EVENT2MASK(DSL_EVENT_S_FIRMWARE_REQUEST);
          
               DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bFirmwareEventAssigned, DSL_TRUE);
            }
         }
         break;
      case DSL_EVENT_S_LINIT_FAILURE:
      case DSL_EVENT_S_LINE_STATE:
      case DSL_EVENT_S_LINE_POWERMANAGEMENT_STATE:
      case DSL_EVENT_S_CHANNEL_DATARATE:
      case DSL_EVENT_S_FIRMWARE_ERROR:
      case DSL_EVENT_S_INIT_READY:
      case DSL_EVENT_S_FE_INVENTORY_AVAILABLE:
      case DSL_EVENT_S_SYSTEM_STATUS:
      case DSL_EVENT_S_PM_SYNC:
      case DSL_EVENT_S_LINE_TRANSMISSION_STATUS:
      case DSL_EVENT_S_SHOWTIME_LOGGING:
      case DSL_EVENT_S_SNMP_MESSAGE_AVAILABLE:
      case DSL_EVENT_S_SYSTEM_INTERFACE_STATUS:
      case DSL_EVENT_S_FIRMWARE_DOWNLOAD_STATUS:
      case DSL_EVENT_S_AUTOBOOT_STATUS:
         if (pData->data.bMask)
         {
            nEventMask |= DSL_EVENT2MASK(pData->data.nEventType);         
         }
         else
         {
            /* Unmask (enable) specified event*/
            nEventMask &= ~DSL_EVENT2MASK(pData->data.nEventType);
         }
         break;
      default:
         DSL_DEBUG(DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - unknown status event(0x%x)"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext), pData->data.nEventType));
         nErrCode = DSL_ERROR;
      break;
      }

      if (nErrCode != DSL_ERROR)
      {
         /* Update event mask*/
         pOpenCtx->nEventMask = nEventMask;         
      }
   }
   else
   {
      nErrCode = DSL_ERR_EVENTS_NOT_ACTIVE;
   }

   DSL_DRV_MUTEX_UNLOCK(pOpenCtx->eventMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_EventStatusMaskConfigSet, "
      "retCode(%d)"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_EVENT_STATUS_MASK_CONFIG_GET
*/
DSL_uint32_t DSL_DRV_EventStatusMaskConfigGet(
   DSL_OpenContext_t *pOpenCtx,
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_EventStatusMask_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint32_t nEventMask;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_EventStatusMaskConfigGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(DSL_DRV_MUTEX_LOCK(pOpenCtx->eventMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: Couldn't lock event mutex"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   /* Check if the event handling is activated for this instance*/
   if (pOpenCtx->bEventActivation)
   {
      /* Get current event mask*/
      nEventMask = pOpenCtx->nEventMask;

      switch(pData->data.nEventType)
      {
         case DSL_EVENT_S_FIRMWARE_REQUEST:
         case DSL_EVENT_S_LINIT_FAILURE:
         case DSL_EVENT_S_LINE_STATE:
         case DSL_EVENT_S_LINE_POWERMANAGEMENT_STATE:
         case DSL_EVENT_S_CHANNEL_DATARATE:
         case DSL_EVENT_S_FIRMWARE_ERROR:
         case DSL_EVENT_S_INIT_READY:
         case DSL_EVENT_S_FE_INVENTORY_AVAILABLE:
         case DSL_EVENT_S_SYSTEM_STATUS:
         case DSL_EVENT_S_PM_SYNC:
         case DSL_EVENT_S_LINE_TRANSMISSION_STATUS:
         case DSL_EVENT_S_SHOWTIME_LOGGING:
         case DSL_EVENT_S_SNMP_MESSAGE_AVAILABLE:
         case DSL_EVENT_S_SYSTEM_INTERFACE_STATUS:
         case DSL_EVENT_S_FIRMWARE_DOWNLOAD_STATUS:
         case DSL_EVENT_S_AUTOBOOT_STATUS:
            pData->data.bMask = nEventMask & DSL_EVENT2MASK(pData->data.nEventType) ?
                                DSL_TRUE : DSL_FALSE;
            break;
      default:
         DSL_DEBUG(DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - unknown status event(0x%x)"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext), pData->data.nEventType));
         nErrCode = DSL_ERROR;
      break;
      }
   }
   else
   {
      nErrCode = DSL_ERR_EVENTS_NOT_ACTIVE;
   }

   DSL_DRV_MUTEX_UNLOCK(pOpenCtx->eventMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_EventStatusMaskConfigSet, "
      "retCode(%d)"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_FramingParameterStatusGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_FramingParameterStatus_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_FramingParameterStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Call device specific stuff*/
   nErrCode = DSL_DRV_DEV_FramingParameterStatusGet(pContext, pData);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_FramingParameterStatusGet, "
      "retCode(%d)"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_DBG_ModuleLevelSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN DSL_DBG_ModuleLevel_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
#ifndef DSL_DEBUG_DISABLE
   DSL_int_t i = 0;
   DSL_boolean_t bModOk = DSL_FALSE, bLvlOk = DSL_FALSE;
   DSL_debugLevelEntry_t *pDbgEntry = DSL_NULL;
   DSL_debugLevelEntry_t *pDbgLvlNames = DSL_NULL;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   /* Check if debug module number is valid */
   pDbgEntry = &DSL_g_dbgLvl[1];
   if ( (pData->data.nDbgModule >= DSL_DBG_LAST_BLOCK) || (pData->data.nDbgModule == 0) )
   {
      bModOk = DSL_FALSE;
      DSL_DEBUG_SET_ERROR(DSL_ERROR);
      DSL_DEBUG( DSL_DBG_ERR, (pContext,
         "DSL[%02d]: DSL_DRV_DBG_ModuleLevelSet - invalid debug module number specified:"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));
      DSL_DEBUG( DSL_DBG_PRN, (pContext,
         "   nDbgModules:" DSL_DRV_CRLF ) );

      for (i = 1; i < DSL_DBG_LAST_BLOCK; i++, pDbgEntry++)
      {
         DSL_DEBUG( DSL_DBG_PRN,
            (pContext, "      %d: %s" DSL_DRV_CRLF , i, pDbgEntry->pcName) );
      }
   }
   else
   {
      bModOk = DSL_TRUE;
   }

   /* Check if debug level is valid */
   pDbgLvlNames = &DSL_g_dbgLvlNames[0];
   for (i = 0; i < DSL_g_dbgLvlNumber; i++, pDbgLvlNames++)
   {
      if (pDbgLvlNames->nDbgLvl == pData->data.nDbgLevel)
      {
         /* debug level found  */
         bLvlOk = DSL_TRUE;
         break;
      }
   }
   if ( (bLvlOk == DSL_FALSE) && (i == DSL_g_dbgLvlNumber) )
   {
      /* debug level NOT valid  */
      DSL_DEBUG_SET_ERROR(DSL_ERROR);
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: DSL_DRV_DBG_ModuleLevelSet - invalid debug level specified:" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      DSL_DEBUG( DSL_DBG_PRN,
         (pContext, "   nDbgLevels:" DSL_DRV_CRLF ) );
      pDbgLvlNames = &DSL_g_dbgLvlNames[0];
      for (i = 0; i < DSL_g_dbgLvlNumber; i++, pDbgLvlNames++)
      {
         DSL_DEBUG( DSL_DBG_PRN,
            (pContext, "      0x%02X: %s" DSL_DRV_CRLF ,
            pDbgLvlNames->nDbgLvl, pDbgLvlNames->pcName) );
      }
   }

   if ( (bModOk == DSL_TRUE) && (bLvlOk == DSL_TRUE) )
   {
      /* Set new debug level for specified module */
      DSL_g_dbgLvl[pData->data.nDbgModule].nDbgLvl = pData->data.nDbgLevel;
      DSL_DEBUG( DSL_DBG_LOCAL,
         (pContext, "DSL[%02d]: new debug level: %s = %s - 0x%02X" DSL_DRV_CRLF ,
              DSL_DEV_NUM(pContext), DSL_g_dbgLvl[pData->data.nDbgModule].pcName,
              pDbgLvlNames->pcName,
              pData->data.nDbgLevel) );
   }
   else
   {
      nErrCode = DSL_ERROR;
   }
#else /* #ifndef DSL_DEBUG_DISABLE */
   /* DSL_DRV_DBG_ModuleLevelSet: Implementation only available if debugging is
      enabled. Please undefine preprocessor directive 'DSL_DEBUG_DISABLE' */
   nErrCode = DSL_ERR_ONLY_SUPPORTED_WITH_DEBUG_ENABLED;
#endif /* #ifndef DSL_DEBUG_DISABLE */

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_DBG_ModuleLevelGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_DBG_ModuleLevel_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
#ifndef DSL_DEBUG_DISABLE
   DSL_boolean_t bLvlOk = DSL_FALSE;
   DSL_debugLevelEntry_t *pDbgEntry = DSL_NULL;
   DSL_debugLevelEntry_t *pDbgLvlNames = DSL_NULL;
   DSL_int_t i = 0;

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   /* Check if debug module number is valid */
   pDbgEntry = &DSL_g_dbgLvl[1];
   if ( (pData->data.nDbgModule >= DSL_DBG_LAST_BLOCK) || (pData->data.nDbgModule == 0) )
   {
      nErrCode = DSL_ERROR;
      DSL_DEBUG_SET_ERROR(nErrCode);
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: DSL_DRV_DBG_ModuleLevelGet - "
         "invalid debug module number specified:" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      DSL_DEBUG( DSL_DBG_PRN,
         (pContext, "   nDbgModules:" DSL_DRV_CRLF ) );
      for (i = 1; i < DSL_DBG_LAST_BLOCK; i++, pDbgEntry++)
      {
         DSL_DEBUG( DSL_DBG_PRN,
            (pContext, "      %d: %s" DSL_DRV_CRLF , i, pDbgEntry->pcName) );
      }
   }
   else
   {
      pData->data.nDbgLevel = DSL_g_dbgLvl[pData->data.nDbgModule].nDbgLvl;

      pDbgLvlNames = &DSL_g_dbgLvlNames[0];
      for (i = 0; i < DSL_g_dbgLvlNumber; i++, pDbgLvlNames++)
      {
         if (pDbgLvlNames->nDbgLvl == pData->data.nDbgLevel)
         {
            /* debug level found  */
            bLvlOk = DSL_TRUE;
            DSL_DEBUG( DSL_DBG_LOCAL,
               (pContext, "DSL[%02d]: actual debug level: %s = %s - 0x%02X" DSL_DRV_CRLF,
                    DSL_DEV_NUM(pContext), DSL_g_dbgLvl[pData->data.nDbgModule].pcName,
                    pDbgLvlNames->pcName,
                    pData->data.nDbgLevel) );
            break;
         }
      }

      if (bLvlOk == DSL_FALSE)
      {
         DSL_DEBUG( DSL_DBG_LOCAL,
            (pContext, "DSL[%02d]: actual debug level: %s = 0x%02X" DSL_DRV_CRLF,
                 DSL_DEV_NUM(pContext), DSL_g_dbgLvl[pData->data.nDbgModule].pcName,
                 pData->data.nDbgLevel) );
      }
   }
#else /* #ifndef DSL_DEBUG_DISABLE */
   /* DSL_DRV_DBG_ModuleLevelGet: Implementation only available if debugging is
      enabled. Please undefine preprocessor directive 'DSL_DEBUG_DISABLE' */
   nErrCode = DSL_ERR_ONLY_SUPPORTED_WITH_DEBUG_ENABLED;
#endif /* #ifndef DSL_DEBUG_DISABLE */

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_DBG_DeviceMessageSend(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_DeviceMessage_t *pMsg)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pMsg);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_DBG_DeviceMessageSend"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   nErrCode = DSL_DRV_DEV_DBG_DeviceMessageSend(pContext, &(pMsg->data));

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_DBG_DeviceMessageSend, retCode(%d)"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_InitDataPrepare(
   DSL_Context_t *pContext,
   DSL_boolean_t bIsInKernel,
   DSL_Init_t *pInit)
{
   DSL_G997_LineInventoryNeData_t *pInv = DSL_NULL;
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pInit);
   DSL_CHECK_ERR_CODE();

   /* Store FW in the DSL CPE internal memory*/
   nErrCode = DSL_DRV_FwStore(
                 pContext, bIsInKernel, pInit->data.pFirmware,
                 pInit->data.nFirmwareSize, pInit->data.pFirmware2,
                 pInit->data.nFirmwareSize2);

   if( nErrCode != DSL_SUCCESS)
   {
      return nErrCode;
   }

   if (pInit->data.pInventory != DSL_NULL)
   {
      pInv = DSL_DRV_PMalloc(sizeof(DSL_G997_LineInventoryNeData_t));
      if (pInv == DSL_NULL)
      {
         nErrCode = DSL_ERR_MEMORY;
      }
      else
      {
         DSL_IoctlMemCpyFrom(bIsInKernel, pInv, pInit->data.pInventory,
            sizeof(DSL_G997_LineInventoryNeData_t));
         pInit->data.pInventory = pInv;
      }
   }

   if (nErrCode == DSL_SUCCESS)
   {
      nErrCode = DSL_DRV_DEV_InitDataPrepare(pContext, bIsInKernel, pInit);
   }

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_void_t DSL_DRV_InitDataFree(
   DSL_Context_t *pContext,
   DSL_Init_t *pInit)
{
   if (pInit == DSL_NULL)
   {
      return;
   }

   if (pInit->data.pInventory != DSL_NULL)
   {
      DSL_DRV_PFree(pInit->data.pInventory);
   }

#ifdef INCLUDE_FW_REQUEST_SUPPORT
   /* Remove stored FW from the DSL CPE context. The FW binary will
       be requested via the corresponding event further if necessary*/
   if( pContext->pFirmware != DSL_NULL )
   {
      DSL_DRV_VFree(pContext->pFirmware);
      pContext->pFirmware     = DSL_NULL;
      pContext->nFirmwareSize = 0;
   }

   if( pContext->pFirmware2 != DSL_NULL )
   {
      DSL_DRV_VFree(pContext->pFirmware2);
      pContext->pFirmware2     = DSL_NULL;
      pContext->nFirmwareSize2 = 0;
   }
#endif /* #ifdef INCLUDE_FW_REQUEST_SUPPORT*/

   DSL_DRV_DEV_InitDataFree(pInit);
}

DSL_Error_t DSL_DRV_IoctlHandleHelperCall(
   DSL_Context_t *pContext,
   DSL_boolean_t bIsInKernel,
   DSL_IoctlHandlerHelperType_t nType,
   DSL_DRV_IoctlHandlerHelperFunc_t pFunc,
   DSL_void_t *pArg,
   DSL_uint32_t nArgSz)
{
   DSL_Error_t nErrCode = DSL_ERR_MEMORY;
   DSL_IOCTL_arg_t *pIOCTL_arg = DSL_NULL;

   if (nType != DSL_IOCTL_HELPER_GET && nType != DSL_IOCTL_HELPER_SET)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Invalid parameter was passed to "
         "the IOCTL helper!"DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));
      return DSL_ERR_INVALID_PARAMETR;
   }

   if (pArg == DSL_NULL || nArgSz == 0)
   {
      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }
   else
   {
      pIOCTL_arg = DSL_DRV_VMalloc(nArgSz);
      if(pIOCTL_arg == DSL_NULL)
      {
         DSL_IoctlMemCpyTo( bIsInKernel, pArg, &nErrCode, sizeof(DSL_Error_t));
         return nErrCode;
      }

      DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, pArg,
         nArgSz);

      nErrCode = pFunc(pContext, pIOCTL_arg);

      if (nErrCode >= DSL_SUCCESS)
      {
         if (nType == DSL_IOCTL_HELPER_GET)
         {
            pIOCTL_arg->accessCtl.nReturn = nErrCode;

            DSL_IoctlMemCpyTo( bIsInKernel, pArg, pIOCTL_arg,
               nArgSz);
         }
         else
         {
            DSL_IoctlMemCpyTo( bIsInKernel, pArg, &nErrCode,
               sizeof(DSL_Error_t));
         }
      }
      else
      {
         DSL_IoctlMemCpyTo( bIsInKernel, pArg, &nErrCode, sizeof(DSL_Error_t));
      }

      DSL_DRV_VFree(pIOCTL_arg);
   }

   return nErrCode;
}

DSL_Error_t DSL_DRV_IoctlInstanceHandleHelperCall(
   DSL_OpenContext_t *pOpenCtx,
   DSL_Context_t *pContext,
   DSL_boolean_t bIsInKernel,
   DSL_IoctlHandlerHelperType_t nType,
   DSL_DRV_IoctlInstanceHandlerHelperFunc_t pFunc,
   DSL_void_t *pArg,
   DSL_uint32_t nArgSz)
{
   DSL_Error_t nErrCode = DSL_ERR_MEMORY;
   DSL_IOCTL_arg_t *pIOCTL_arg = DSL_NULL;

   if (nType != DSL_IOCTL_HELPER_GET && nType != DSL_IOCTL_HELPER_SET)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Invalid parameter was passed to "
         "the IOCTL helper!"DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));
      return DSL_ERR_INVALID_PARAMETR;
   }

   if (pArg == DSL_NULL || nArgSz == 0)
   {
      nErrCode = DSL_ERR_INVALID_PARAMETR;
   }
   else
   {
      pIOCTL_arg = DSL_DRV_VMalloc(nArgSz);
      if(pIOCTL_arg == DSL_NULL)
      {
         DSL_IoctlMemCpyTo( bIsInKernel, pArg, &nErrCode, sizeof(DSL_Error_t));
         return nErrCode;
      }

      DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, pArg,
         nArgSz);

      nErrCode = pFunc(pOpenCtx, pContext, pIOCTL_arg);

      if (nErrCode >= DSL_SUCCESS)
      {
         if (nType == DSL_IOCTL_HELPER_GET)
         {
            pIOCTL_arg->accessCtl.nReturn = nErrCode;

            DSL_IoctlMemCpyTo( bIsInKernel, pArg, pIOCTL_arg,
               nArgSz);
         }
         else
         {
            DSL_IoctlMemCpyTo( bIsInKernel, pArg, &nErrCode,
               sizeof(DSL_Error_t));
         }
      }
      else
      {
         DSL_IoctlMemCpyTo( bIsInKernel, pArg, &nErrCode, sizeof(DSL_Error_t));
      }

      DSL_DRV_VFree(pIOCTL_arg);
   }

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_IoctlHandle(
   DSL_OpenContext_t *pOpenCtx,
   DSL_Context_t *pContext,
   DSL_boolean_t bIsInKernel,
   DSL_uint_t nCommand,
   DSL_uint32_t nArg)
{
   DSL_Error_t nErrCode = DSL_ERROR;
   DSL_IOCTL_arg_t *pIOCTL_arg = DSL_NULL;
   DSL_EventData_Union_t *pEventData;
   DSL_uint8_t *pMsgBuf, *pOldMsgBuf;
#ifdef INCLUDE_DSL_CPE_TRACE_BUFFER
   DSL_uint16_t i; /* FOR DEBUG */
#endif
   DSL_boolean_t bEventActivation = DSL_FALSE;   
   DSL_BF_ResourceActivationType_t nResourceActivationMask;
   DSL_EventStatusData_t event;

   pIOCTL_arg = (DSL_IOCTL_arg_t*)DSL_DRV_VMalloc(sizeof(DSL_IOCTL_arg_t));
   if(pIOCTL_arg == DSL_NULL)
   {
      DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, &nErrCode,
         sizeof(DSL_Error_t));
      return nErrCode;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_IoctlHandle, call %d - "
      "(%s)"DSL_DRV_CRLF, DSL_DEV_NUM(pContext),
      nCommand, DSL_DBG_IoctlName(nCommand)));

   switch (nCommand)
   {
      case DSL_FIO_INIT:
         if (nArg == 0)
         {
            nErrCode = DSL_ERR_INVALID_PARAMETR;
         }
         else
         {
            DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, (DSL_void_t*)nArg,
                                 sizeof(DSL_Init_t));

            nErrCode = DSL_DRV_InitDataPrepare(pContext, bIsInKernel, &(pIOCTL_arg->init));
            if (nErrCode == DSL_SUCCESS)
            {
               nErrCode = DSL_DRV_Init(pContext, &(pIOCTL_arg->init));
            }

            DSL_DRV_InitDataFree(pContext, &(pIOCTL_arg->init));

            DSL_DEBUG(DSL_DBG_MSG,
               (pContext, "DSL[%02d]: DSL_DRV_IoctlHandle - return"
               "(from DSL_FIO_INIT) %d"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), nErrCode));
            DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, &nErrCode,
               sizeof(DSL_Error_t));
         }
         break;

      case DSL_FIO_AUTOBOOT_LOAD_FIRMWARE:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_AutobootLoadFirmware,
            (DSL_void_t*) nArg, sizeof(DSL_AutobootLoadFirmware_t));
         break;

      case DSL_FIO_AUTOBOOT_CONTROL_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_AutobootControlSet,
            (DSL_void_t*) nArg, sizeof(DSL_AutobootControl_t));
         break;

      case DSL_FIO_AUTOBOOT_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_AutobootStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_AutobootStatus_t));
         break;

      case DSL_FIO_AUTOBOOT_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_AutobootConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_AutobootConfig_t));
         break;

      case DSL_FIO_AUTOBOOT_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_AutobootConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_AutobootConfig_t));
         break;

      case DSL_FIO_LINE_STATE_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LineStateGet,
            (DSL_void_t*) nArg, sizeof(DSL_LineState_t));
         break;

      case DSL_FIO_LINE_FEATURE_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LineFeatureConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_LineFeature_t));
         break;

      case DSL_FIO_LINE_FEATURE_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LineFeatureConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_LineFeature_t));
         break;

      case DSL_FIO_LINE_FEATURE_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LineFeatureStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_LineFeature_t));
         break;

      case DSL_FIO_INTEROP_FEATURE_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_InteropFeatureConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_InteropFeatureConfig_t));
         break;

      case DSL_FIO_INTEROP_FEATURE_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_InteropFeatureConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_InteropFeatureConfig_t));
         break;

      case DSL_FIO_DBG_MODULE_LEVEL_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_DBG_ModuleLevelSet,
            (DSL_void_t*) nArg, sizeof(DSL_DBG_ModuleLevel_t));
         break;

      case DSL_FIO_DBG_MODULE_LEVEL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_DBG_ModuleLevelGet,
            (DSL_void_t*) nArg, sizeof(DSL_DBG_ModuleLevel_t));
         break;

      case DSL_FIO_VERSION_INFORMATION_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_VersionInformationGet,
            (DSL_void_t*) nArg, sizeof(DSL_VersionInformation_t));
         break;

      case DSL_FIO_EVENT_STATUS_MASK_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlInstanceHandleHelperCall(
            pOpenCtx, pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlInstanceHandlerHelperFunc_t)DSL_DRV_EventStatusMaskConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_EventStatusMask_t));
         break;

      case DSL_FIO_EVENT_STATUS_MASK_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlInstanceHandleHelperCall(
            pOpenCtx, pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlInstanceHandlerHelperFunc_t)DSL_DRV_EventStatusMaskConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_EventStatusMask_t));
         break;

      case DSL_FIO_SYSTEM_INTERFACE_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SystemInterfaceConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_SystemInterfaceConfig_t));
         break;

      case DSL_FIO_SYSTEM_INTERFACE_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SystemInterfaceConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_SystemInterfaceConfig_t));
         break;

      case DSL_FIO_SYSTEM_INTERFACE_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SystemInterfaceStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_SystemInterfaceStatus_t));
         break;

#if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)
      case DSL_FIO_LINE_PATH_COUNTER_TOTAL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LinePathCounterTotalGet,
            (DSL_void_t*) nArg, sizeof(DSL_LinePathCounterTotal_t));
         break;

      case DSL_FIO_DATA_PATH_COUNTER_TOTAL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_DataPathCounterTotalGet,
            (DSL_void_t*) nArg, sizeof(DSL_DataPathCounterTotal_t));
         break;
         
      case DSL_FIO_LOW_LEVEL_CONFIGURATION_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LowLevelConfigurationSet,
            (DSL_void_t*) nArg, sizeof(DSL_LowLevelConfiguration_t));
         break;

      case DSL_FIO_LOW_LEVEL_CONFIGURATION_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_LowLevelConfigurationGet,
            (DSL_void_t*) nArg, sizeof(DSL_LowLevelConfiguration_t));
         break;

      case DSL_FIO_AUX_LINE_INVENTORY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_AuxLineInventoryGet,
            (DSL_void_t*) nArg, sizeof(DSL_AuxLineInventory_t));
         break;

      case DSL_FIO_BAND_PLAN_SUPPORT_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_BandPlanSupportedGet,
            (DSL_void_t*) nArg, sizeof(DSL_BandPlanSupport_t));
         break;

      case DSL_FIO_BAND_PLAN_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_BandPlanStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_BandPlanStatus_t));
         break;

      case DSL_FIO_EFM_MAC_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_EfmMacConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_EFM_MacConfig_t));
         break;

      case DSL_FIO_EFM_MAC_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_EfmMacConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_EFM_MacConfig_t));
         break;

      /* SAR block*/
      case DSL_FIO_SAR_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_ConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_Config_t));
         break;

      case DSL_FIO_SAR_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_ConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_Config_t));
         break;

      case DSL_FIO_SAR_SEGMENTATION_TABLE_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_SegmentationTableSet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_SegmentationTable_t));
         break;

      case DSL_FIO_SAR_SEGMENTATION_TABLE_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_SegmentationTableGet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_SegmentationTable_t));
         break;

      case DSL_FIO_SAR_REASSEMBLY_TABLE_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_ReassemblyTableSet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_ReassemblyTable_t));
         break;

      case DSL_FIO_SAR_REASSEMBLY_TABLE_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_ReassemblyTableGet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_ReassemblyTable_t));
         break;

      case DSL_FIO_SAR_SEGMENTATION_COUNTERS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_SegmentationCountersGet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_SegmentationCounters_t));
         break;

      case DSL_FIO_SAR_REASSEMBLY_COUNTERS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_SAR_ReassemblyCountersGet,
            (DSL_void_t*) nArg, sizeof(DSL_SAR_ReassemblyCounters_t));
         break;

      case DSL_FIO_UTOPIA_BUS_WIDTH_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_UtopiaBusWidthConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_UtopiaBusWidthConfig_t));
         break;

      case DSL_FIO_UTOPIA_BUS_WIDTH_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_UtopiaBusWidthConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_UtopiaBusWidthConfig_t));
         break;
#endif /* #if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)*/

      case DSL_FIO_FRAMING_PARAMETER_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_FramingParameterStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_FramingParameterStatus_t));
         break;

      case DSL_FIO_UTOPIA_ADDRESS_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_UtopiaAddressConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_PhyAddressConfig_t));
         break;

      case DSL_FIO_UTOPIA_ADDRESS_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_UtopiaAddressConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_PhyAddressConfig_t));
         break;

      case DSL_FIO_TEST_MODE_CONTROL_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_TestModeControlSet,
            (DSL_void_t*) nArg, sizeof(DSL_TestModeControl_t));
         break;

      case DSL_FIO_TEST_MODE_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_TestModeStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_TestModeStatus_t));
         break;

#ifdef INCLUDE_DSL_CPE_TRACE_BUFFER
      case DSL_FIO_SHOWTIME_LOGGING_DATA_GET:
         if (nArg == 0)
         {
            nErrCode = DSL_ERR_INVALID_PARAMETR;
         }
         else
         {
            DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, (DSL_void_t*)nArg,
                                 sizeof(DSL_ShowtimeLogging_t));

            DSL_DEBUG(DSL_DBG_MSG, (pContext,
               "DSL[%02d]: DEBUG:2:: showtime event logging data (in the kernel space):",
               DSL_DEV_NUM(pContext)));
            for (i = 0; i < DSL_DEV_SHOWTIME_EVENT_LOGGING_BUFFER_LENGTH; i++)
            {
               if (i % 10 == 0) DSL_DEBUG(DSL_DBG_MSG, (pContext, DSL_DRV_CRLF));
               DSL_DEBUG(DSL_DBG_MSG, (pContext, "0x%04X ", pContext->loggingBuffer[i]));
            }
            DSL_DEBUG(DSL_DBG_MSG, (pContext, DSL_DRV_CRLF DSL_DRV_CRLF));

            if (pIOCTL_arg->showtimeLogging.data.pData == DSL_NULL)
            {
               /* in this case return data size only */
               pIOCTL_arg->showtimeLogging.data.nDataSize =
                  DSL_DEV_SHOWTIME_EVENT_LOGGING_BUFFER_LENGTH * 2;
               nErrCode = DSL_SUCCESS;
            }
            else
            {
               if ((pIOCTL_arg->showtimeLogging.data.nDataOffset >=
                  DSL_DEV_SHOWTIME_EVENT_LOGGING_BUFFER_LENGTH * 2) ||
                  pIOCTL_arg->showtimeLogging.data.nDataSize == 0)
               {
                  nErrCode = DSL_ERR_INVALID_PARAMETR;
               }
               else
               {
                  if (pIOCTL_arg->showtimeLogging.data.nDataSize >=
                      (DSL_DEV_SHOWTIME_EVENT_LOGGING_BUFFER_LENGTH * 2 -
                       pIOCTL_arg->showtimeLogging.data.nDataOffset))
                  {
                     DSL_IoctlMemCpyTo(bIsInKernel,
                        pIOCTL_arg->showtimeLogging.data.pData,
                        pContext->loggingBuffer,
                        DSL_DEV_SHOWTIME_EVENT_LOGGING_BUFFER_LENGTH * 2 -
                           pIOCTL_arg->showtimeLogging.data.nDataOffset);
                     pIOCTL_arg->showtimeLogging.data.nDataOffset = 0;
                     pIOCTL_arg->showtimeLogging.data.nDataSize =
                        DSL_DEV_SHOWTIME_EVENT_LOGGING_BUFFER_LENGTH * 2 -
                        pIOCTL_arg->showtimeLogging.data.nDataOffset;
                  }
                  else
                  {
                     DSL_IoctlMemCpyTo(bIsInKernel,
                        pIOCTL_arg->showtimeLogging.data.pData,
                        pContext->loggingBuffer,
                        pIOCTL_arg->showtimeLogging.data.nDataSize);
                     pIOCTL_arg->showtimeLogging.data.nDataOffset +=
                        pIOCTL_arg->showtimeLogging.data.nDataSize;
                  }
                  nErrCode = DSL_SUCCESS;
               }
            }

            DSL_DEBUG(DSL_DBG_MSG, (pContext,
               "DSL[%02d]: DEBUG:3:: showtime event logging data (in the user space):",
               DSL_DEV_NUM(pContext)));
            for (i = 0; i < pIOCTL_arg->showtimeLogging.data.nDataSize / 2; i++)
            {
               if (i % 10 == 0) DSL_DEBUG(DSL_DBG_MSG, (pContext, DSL_DRV_CRLF));
               DSL_DEBUG(DSL_DBG_MSG, (pContext, "0x%04X ",
                  ((DSL_uint16_t*)(pIOCTL_arg->showtimeLogging.data.pData))[i]));
            }
            DSL_DEBUG(DSL_DBG_MSG, (pContext, DSL_DRV_CRLF DSL_DRV_CRLF));

            pIOCTL_arg->showtimeLogging.accessCtl.nReturn = nErrCode;

            if (nErrCode == DSL_SUCCESS)
            {
               DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, pIOCTL_arg,
                  sizeof(DSL_ShowtimeLogging_t));
            }
            else
            {
               DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, &nErrCode,
                  sizeof(DSL_Error_t));
            }
         }
         break;
#endif

      case DSL_FIO_DBG_DEVICE_MESSAGE_SEND:
         if (nArg == 0)
         {
            nErrCode = DSL_ERR_INVALID_PARAMETR;
         }
         else
         {
            DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, (DSL_void_t*)nArg,
               sizeof(DSL_DeviceMessage_t));

            if ((pOldMsgBuf = pIOCTL_arg->dbgMsg.data.pMsg) == DSL_NULL)
            {
               nErrCode = DSL_ERR_INVALID_PARAMETR;
            }
            else
            {
               pMsgBuf = DSL_DRV_VMalloc(
                  (pIOCTL_arg->dbgMsg.data.nSizeRx
                     > pIOCTL_arg->dbgMsg.data.nSizeTx) ?
                     pIOCTL_arg->dbgMsg.data.nSizeRx :
                     pIOCTL_arg->dbgMsg.data.nSizeTx);
               if (pMsgBuf == DSL_NULL)
               {
                  nErrCode = DSL_ERR_MEMORY;
               }
               else
               {
                  DSL_IoctlMemCpyFrom(bIsInKernel, pMsgBuf, pOldMsgBuf,
                     pIOCTL_arg->dbgMsg.data.nSizeTx);

                  pIOCTL_arg->dbgMsg.data.pMsg = pMsgBuf;
                  nErrCode = DSL_DRV_DBG_DeviceMessageSend(pContext,
                     &pIOCTL_arg->dbgMsg);

                  if (nErrCode == DSL_SUCCESS)
                  {
                     if (pIOCTL_arg->dbgMsg.data.nSizeRx != 0)
                     {
                        DSL_IoctlMemCpyTo(bIsInKernel, pOldMsgBuf, pMsgBuf,
                           pIOCTL_arg->dbgMsg.data.nSizeRx);
                     }
                     pIOCTL_arg->dbgMsg.data.pMsg = pOldMsgBuf;
                     DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, pIOCTL_arg,
                        sizeof(DSL_DeviceMessage_t));
                  }
                  DSL_DRV_VFree(pMsgBuf);
               }
            }

            DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, &nErrCode,
               sizeof(DSL_Error_t));
         }
         break;

      case DSL_FIO_EVENT_STATUS_GET:
         if (nArg == 0)
         {
            nErrCode = DSL_ERR_INVALID_PARAMETR;
         }
         else
         {
            DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, (DSL_void_t*)nArg,
               sizeof(DSL_EventStatus_t));

            pEventData = pIOCTL_arg->event.data.pData;

            nErrCode = DSL_DRV_EventUnqueue(pOpenCtx, &pIOCTL_arg->event.data);

            pIOCTL_arg->event.accessCtl.nReturn = nErrCode;

            if (nErrCode == DSL_SUCCESS)
            {
               if (pEventData != DSL_NULL && pIOCTL_arg->event.data.pData
                  != DSL_NULL)
               {
                  DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)pEventData,
                     pIOCTL_arg->event.data.pData,
                     sizeof(DSL_EventData_Union_t));
               }

               if (pIOCTL_arg->event.data.pData != DSL_NULL)
               {
                  DSL_DRV_VFree(pIOCTL_arg->event.data.pData);
                  pIOCTL_arg->event.data.pData = pEventData;
               }
               else
               {
                  pIOCTL_arg->event.data.pData = DSL_NULL;
               }

               DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, pIOCTL_arg,
                  sizeof(DSL_EventStatus_t));
            }
            else
            {
               DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, &nErrCode,
                  sizeof(DSL_Error_t));
            }
         }

         break;

      case DSL_FIO_INSTANCE_CONTROL_SET:
         if (nArg == 0)
         {
            nErrCode = DSL_ERR_INVALID_PARAMETR;
         }
         else
         {
            DSL_IoctlMemCpyFrom(bIsInKernel, pIOCTL_arg, (DSL_void_t*)nArg,
               sizeof(DSL_InstanceControl_t));

            nErrCode = DSL_SUCCESS;

            /* Get bEventActivation flag*/
            bEventActivation = (((DSL_InstanceControl_t*)nArg)->data.bEventActivation);
            /* Get nResourceActivationMask data*/
            nResourceActivationMask = (((DSL_InstanceControl_t*)nArg)->data.nResourceActivationMask);

            /* 
               Proceed Event Activation Contriol settings
            */
            if(DSL_DRV_MUTEX_LOCK(pOpenCtx->eventMutex))
            {
               DSL_DEBUG( DSL_DBG_ERR,
                  (pContext, "DSL[%02d]: Couldn't lock event mutex"DSL_DRV_CRLF,
                  DSL_DEV_NUM(pContext)));
               nErrCode = DSL_ERROR;
            }
            else
            {
               if (bEventActivation == DSL_FALSE)
               {
                  DSL_DRV_MUTEX_UNLOCK(pOpenCtx->eventMutex);
                  /** free event structures */
                  if (pOpenCtx->eventFifo != DSL_NULL && pOpenCtx->eventFifoBuf != DSL_NULL)
                  {
                     while (!DSL_Fifo_isEmpty(pOpenCtx->eventFifo))
                     {
                        if (DSL_DRV_EventUnqueue(pOpenCtx, &event) == DSL_SUCCESS)
                        {
                           if (event.pData != DSL_NULL)
                           {
                              DSL_DRV_VFree(event.pData);
                           }
                        }
                     }
                  }

                  if(DSL_DRV_MUTEX_LOCK(pOpenCtx->eventMutex))
                  {
                     DSL_DEBUG( DSL_DBG_ERR,
                        (pContext, "DSL[%02d]: Couldn't lock event mutex"DSL_DRV_CRLF,
                        DSL_DEV_NUM(pContext)));
                     nErrCode = DSL_ERROR;
                  }
                  else
                  {
                     /* release event structures */
                     if (pOpenCtx->eventFifo != DSL_NULL)
                     {
                        DSL_DRV_VFree(pOpenCtx->eventFifo);
                        pOpenCtx->eventFifo = DSL_NULL;
                     }

                     /* release event structures */
                     if (pOpenCtx->eventFifoBuf != DSL_NULL)
                     {
                        DSL_DRV_VFree(pOpenCtx->eventFifoBuf);
                        pOpenCtx->eventFifoBuf = DSL_NULL;
                     }
                  }
               }
               else
               {
                  /* initialize event structures */
                  pOpenCtx->eventFifo = DSL_DRV_VMalloc(sizeof(DSL_FIFO));
                  if (pOpenCtx->eventFifo == DSL_NULL)
                  {
                     DSL_DEBUG(DSL_DBG_ERR,
                        (DSL_NULL, "DSL[%02d]: Open failed, "
                        "could not allocate memory for event FIFO..."DSL_DRV_CRLF,
                        DSL_DEV_NUM(pContext)));
                     nErrCode = DSL_ERR_MEMORY;
                  }
                  else
                  {
                     pOpenCtx->eventFifoBuf = DSL_DRV_VMalloc(DSL_EVENT_FIFO_ELEMENT_COUNT
                        * sizeof(DSL_EventStatusData_t));
                     if (pOpenCtx->eventFifoBuf == DSL_NULL)
                     {
                        DSL_DEBUG(DSL_DBG_ERR,
                           (DSL_NULL, "DSL[%02d]: Open failed, "
                           "could not allocate memory for event FIFO..."DSL_DRV_CRLF,
                           DSL_DEV_NUM(pContext)));
                        DSL_DRV_VFree(pOpenCtx->eventFifo);
                        nErrCode = DSL_ERR_MEMORY;
                     }
                     else
                     {
                        /* init event fifo */
                        if (DSL_Fifo_Init(pOpenCtx->eventFifo,
                           pOpenCtx->eventFifoBuf,
                           pOpenCtx->eventFifoBuf +
                             (DSL_EVENT_FIFO_ELEMENT_COUNT - 1) *
                             sizeof(DSL_EventStatusData_t),
                           sizeof(DSL_EventStatusData_t)) != 0)
                        {
                           DSL_DEBUG(DSL_DBG_ERR,
                              (DSL_NULL, "DSL[%02d]: Internal error in the DSL "
                              "CPE API. Could not prepare FIFO for the events."DSL_DRV_CRLF,
                              DSL_DEV_NUM(pContext)));
                           DSL_DRV_VFree(pOpenCtx->eventFifoBuf);
                           DSL_DRV_VFree(pOpenCtx->eventFifo);
                           nErrCode = DSL_ERR_INTERNAL;
                        }
                     }
                  }
               }

               pOpenCtx->bEventActivation = bEventActivation;
               pOpenCtx->nResourceActivationMask = nResourceActivationMask;
               DSL_DRV_MUTEX_UNLOCK(pOpenCtx->eventMutex);
            }

            /* 
               Proceed Resource activation settings settings
            */
#ifdef INCLUDE_DSL_CEOC
            if(DSL_DRV_MUTEX_LOCK(pOpenCtx->rxSnmpFifoMutex))
            {
               DSL_DEBUG( DSL_DBG_ERR,
                  (pContext, "DSL[%02d]: Couldn't lock Rx SNMP FIFO mutex"DSL_DRV_CRLF,
                  DSL_DEV_NUM(pContext)));
               nErrCode = DSL_ERR_SEMAPHORE_GET;
            }
            else
            {
               if (nResourceActivationMask & DSL_RESOURCE_ACTIVATION_SNMP)
               {
                  /* release event structures */
                  if (pOpenCtx->rxSnmpFifo != DSL_NULL)
                  {
                     DSL_DRV_VFree(pOpenCtx->rxSnmpFifo);
                     pOpenCtx->rxSnmpFifo = DSL_NULL;
                  }

                  /* release event structures */
                  if (pOpenCtx->rxSnmpFifoBuf != DSL_NULL)
                  {
                     DSL_DRV_VFree(pOpenCtx->rxSnmpFifoBuf);
                     pOpenCtx->rxSnmpFifoBuf = DSL_NULL;
                  }
               }
               else
               {
                  /* initialize event structures */
                  pOpenCtx->rxSnmpFifo = DSL_DRV_VMalloc(sizeof(DSL_FIFO));
                  if (pOpenCtx->rxSnmpFifo == DSL_NULL)
                  {
                     DSL_DEBUG(DSL_DBG_ERR,
                        (DSL_NULL, "DSL[%02d]: Could not allocate memory for SNMP FIFO!!!"DSL_DRV_CRLF,
                        DSL_DEV_NUM(pContext)));
                     nErrCode = DSL_ERR_MEMORY;
                  }
                  else
                  {
                     pOpenCtx->rxSnmpFifoBuf =
                        DSL_DRV_VMalloc(DSL_CEOC_RX_FIFO_ELEMENT_COUNT * sizeof(DSL_G997_SnmpData_t));
                     if (pOpenCtx->rxSnmpFifoBuf == DSL_NULL)
                     {
                        DSL_DEBUG(DSL_DBG_ERR,
                           (DSL_NULL, "DSL[%02d]: Could not allocate memory for SNMP FIFO buffer!!!"DSL_DRV_CRLF,
                           DSL_DEV_NUM(pContext)));
                        DSL_DRV_VFree(pOpenCtx->rxSnmpFifo);
                        nErrCode = DSL_ERR_MEMORY;
                     }
                     else
                     {
                        /* init event fifo */
                        if (DSL_Fifo_Init(
                               pOpenCtx->rxSnmpFifo,
                               pOpenCtx->rxSnmpFifoBuf,
                               pOpenCtx->rxSnmpFifoBuf +
                               sizeof(DSL_CEOC_Message_t)*(DSL_CEOC_RX_FIFO_ELEMENT_COUNT - 1),
                               sizeof(DSL_CEOC_Message_t)) != 0)
                        {
                           DSL_DEBUG(DSL_DBG_ERR,
                              (DSL_NULL, "DSL[%02d]: Internal error in the DSL "
                              "CPE API. Could not prepare FIFO for the events."DSL_DRV_CRLF,
                              DSL_DEV_NUM(pContext)));
                           DSL_DRV_VFree(pOpenCtx->rxSnmpFifoBuf);
                           DSL_DRV_VFree(pOpenCtx->rxSnmpFifo);
                           nErrCode = DSL_ERR_INTERNAL;
                        }
                     }
                  }               
               }
               pOpenCtx->nResourceActivationMask = nResourceActivationMask;
               DSL_DRV_MUTEX_UNLOCK(pOpenCtx->rxSnmpFifoMutex);
            }
#endif /* #ifdef INCLUDE_DSL_CEOC*/

            DSL_IoctlMemCpyTo( bIsInKernel, (DSL_void_t*)nArg, &nErrCode,
               sizeof(DSL_Error_t));
         }
         break;

      case DSL_FIO_INSTANCE_STATUS_GET:
         nErrCode = DSL_DRV_IoctlInstanceHandleHelperCall(
            pOpenCtx, pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlInstanceHandlerHelperFunc_t)DSL_DRV_InstanceStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_InstanceStatus_t));
         break;

/* G997 block */
#if defined(INCLUDE_DSL_CEOC)
      case DSL_FIO_G997_SNMP_MESSAGE_SEND:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_SnmpMessageSend,
            (DSL_void_t*) nArg, sizeof(DSL_G997_Snmp_t));
         break;
      case DSL_FIO_G997_SNMP_MESSAGE_RECEIVE:
         nErrCode = DSL_DRV_IoctlInstanceHandleHelperCall(
            pOpenCtx, pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlInstanceHandlerHelperFunc_t)DSL_DRV_G997_SnmpMessageReceive,
            (DSL_void_t*) nArg, sizeof(DSL_G997_Snmp_t));
         break;
#endif /* #if defined(INCLUDE_DSL_CEOC)*/

      case DSL_FIO_G997_POWER_MANAGEMENT_STATE_FORCED_TRIGGER:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_PowerManagementStateForcedTrigger,
            (DSL_void_t*) nArg, sizeof(DSL_G997_PowerManagementStateForcedTrigger_t));
         break;

      case DSL_FIO_G997_POWER_MANAGEMENT_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_PowerManagementStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_PowerManagementStatus_t));
         break;

      case DSL_FIO_G997_FRAMING_PARAMETER_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_FramingParameterStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_FramingParameterStatus_t));
         break;

      case DSL_FIO_G997_ALARM_MASK_DATA_PATH_FAILURES_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlInstanceHandleHelperCall(
            pOpenCtx, pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlInstanceHandlerHelperFunc_t)DSL_DRV_G997_AlarmMaskDataPathFailuresConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DataPathFailures_t));
         break;

      case DSL_FIO_G997_ALARM_MASK_DATA_PATH_FAILURES_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_AlarmMaskDataPathFailuresConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DataPathFailures_t));
         break;

      case DSL_FIO_G997_ALARM_MASK_LINE_FAILURES_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlInstanceHandleHelperCall(
            pOpenCtx, pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlInstanceHandlerHelperFunc_t)DSL_DRV_G997_AlarmMaskLineFailuresConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineFailures_t));
         break;

      case DSL_FIO_G997_ALARM_MASK_LINE_FAILURES_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_AlarmMaskLineFailuresConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineFailures_t));
         break;

      case DSL_FIO_G997_LINE_INIT_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineInitStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineInitStatus_t));
         break;

      case DSL_FIO_G997_LINE_FAILURES_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineFailuresStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineFailures_t));
         break;

      case DSL_FIO_G997_XTU_SYSTEM_ENABLING_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_XTUSystemEnablingConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_XTUSystemEnabling_t));
         break;

      case DSL_FIO_G997_XTU_SYSTEM_ENABLING_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_XTUSystemEnablingConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_XTUSystemEnabling_t));
         break;

      case DSL_FIO_G997_XTU_SYSTEM_ENABLING_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_XTUSystemEnablingStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_XTUSystemEnabling_t));
         break;

      case DSL_FIO_G997_CHANNEL_DATA_RATE_THRESHOLD_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_ChannelDataRateThresholdConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_ChannelDataRateThreshold_t));
         break;

      case DSL_FIO_G997_CHANNEL_DATA_RATE_THRESHOLD_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_ChannelDataRateThresholdConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_ChannelDataRateThreshold_t));
         break;

      case DSL_FIO_G997_LINE_TRANSMISSION_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineTransmissionStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineTransmissionStatus_t));
         break;

      case DSL_FIO_G997_LAST_STATE_TRANSMITTED_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LastStateTransmittedGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LastStateTransmitted_t));
         break;

      case DSL_FIO_G997_BIT_ALLOCATION_NSC_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_BitAllocationNSCGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_BitAllocationNsc_t));
         break;

      case DSL_FIO_G997_SNR_ALLOCATION_NSC_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_SnrAllocationNscGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_SnrAllocationNsc_t));
         break;

      case DSL_FIO_G997_GAIN_ALLOCATION_NSC_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_GainAllocationNscGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_GainAllocationNsc_t));
         break;

      case DSL_FIO_G997_LINE_INVENTORY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineInventoryGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineInventory_t));
         break;

      case DSL_FIO_G997_LINE_INVENTORY_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineInventorySet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineInventoryNe_t));
         break;

      case DSL_FIO_G997_LINE_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineStatus_t));
         break;

      case DSL_FIO_G997_LINE_ACTIVATE_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineActivateConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineActivate_t));
         break;

      case DSL_FIO_G997_LINE_ACTIVATE_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineActivateConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineActivate_t));
         break;

#if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)
      case DSL_FIO_G997_LINE_STATUS_PER_BAND_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_LineStatusPerBandGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_LineStatusPerBand_t));
         break;

      case DSL_FIO_G997_RATE_ADAPTATION_CONFIG_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_RateAdaptationConfigSet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_RateAdaptationConfig_t));
         break;

      case DSL_FIO_G997_RATE_ADAPTATION_CONFIG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_RateAdaptationConfigGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_RateAdaptationConfig_t));
         break;
#endif /* (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1) */

      case DSL_FIO_G997_CHANNEL_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_ChannelStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_ChannelStatus_t));
         break;

      case DSL_FIO_G997_DATA_PATH_FAILURES_STATUS_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DataPathFailuresStatusGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DataPathFailures_t));
         break;

      case DSL_FIO_G997_DELT_HLIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DeltHlinGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DeltHlin_t));
         break;

      case DSL_FIO_G997_DELT_HLIN_SCALE_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DeltHlinScaleGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DeltHlinScale_t));
         break;

      case DSL_FIO_G997_DELT_HLOG_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DeltHlogGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DeltHlog_t));
         break;

      case DSL_FIO_G997_DELT_QLN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DeltQLNGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DeltQln_t));
         break;

      case DSL_FIO_G997_DELT_SNR_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DeltSNRGet,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DeltSnr_t));
         break;

      case DSL_FIO_G997_DELT_FREE_RESOURCES:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_G997_DeltFreeResources,
            (DSL_void_t*) nArg, sizeof(DSL_G997_DeltFreeResources_t));
         break;

/* PM block */
#if defined(INCLUDE_DSL_PM)
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_ELAPSED_TIME_RESET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ElapsedTimeReset,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ElapsedTimeReset_t));
         break;

      case DSL_FIO_PM_BURNIN_MODE_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_BurninModeSet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_BurninMode_t));
         break;

      case DSL_FIO_PM_SYNC_MODE_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_SyncModeSet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_SyncMode_t));
         break;

      case DSL_FIO_PM_15MIN_ELAPSED_EXT_TRIGGER:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_15MinElapsedExtTrigger,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ElapsedExtTrigger_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
      case DSL_FIO_PM_RESET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_Reset,
            (DSL_void_t*) nArg, sizeof(DSL_PM_Reset_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
      case DSL_FIO_PM_CHANNEL_COUNTERS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelCounters15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelCounters_t));
         break;

      case DSL_FIO_PM_CHANNEL_COUNTERS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelCounters1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
      case DSL_FIO_PM_CHANNEL_COUNTERS_TOTAL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelCountersTotalGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelCountersTotal_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
      case DSL_FIO_PM_CHANNEL_COUNTERS_SHOWTIME_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelCountersShowtimeGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_CHANNEL_HISTORY_STATS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelHistoryStats15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStatsChDir_t));
         break;

      case DSL_FIO_PM_CHANNEL_HISTORY_STATS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelHistoryStats1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStatsChDir_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#ifdef INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_CHANNEL_THRESHOLDS_15MIN_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelThresholds15MinSet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelThreshold_t));
         break;

      case DSL_FIO_PM_CHANNEL_THRESHOLDS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelThresholds15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelThreshold_t));
         break;

      case DSL_FIO_PM_CHANNEL_THRESHOLDS_1DAY_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelThresholds1DaySet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelThreshold_t));
         break;

      case DSL_FIO_PM_CHANNEL_THRESHOLDS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_ChannelThresholds1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_ChannelThreshold_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS*/
#endif /* # ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
      case DSL_FIO_PM_DATA_PATH_COUNTERS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathCounters15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathCounters_t));
         break;

      case DSL_FIO_PM_DATA_PATH_COUNTERS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathCounters1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
      case DSL_FIO_PM_DATA_PATH_COUNTERS_TOTAL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathCountersTotalGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathCountersTotal_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
      case DSL_FIO_PM_DATA_PATH_COUNTERS_SHOWTIME_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathCountersShowtimeGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_DATA_PATH_HISTORY_STATS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathHistoryStats15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStatsChDir_t));
         break;

      case DSL_FIO_PM_DATA_PATH_HISTORY_STATS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathHistoryStats1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStatsChDir_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_DATA_PATH_THRESHOLDS_15MIN_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathThresholds15MinSet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathThreshold_t));
         break;

      case DSL_FIO_PM_DATA_PATH_THRESHOLDS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathThresholds15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathThreshold_t));
         break;

      case DSL_FIO_PM_DATA_PATH_THRESHOLDS_1DAY_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathThresholds1DaySet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathThreshold_t));
         break;

      case DSL_FIO_PM_DATA_PATH_THRESHOLDS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_DataPathThresholds1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_DataPathThreshold_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
      case DSL_FIO_PM_LINE_SEC_COUNTERS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecCounters15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecCounters_t));
         break;

      case DSL_FIO_PM_LINE_SEC_COUNTERS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecCounters1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
      case DSL_FIO_PM_LINE_SEC_COUNTERS_TOTAL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecCountersTotalGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecCountersTotal_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
      case DSL_FIO_PM_LINE_SEC_COUNTERS_SHOWTIME_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecCountersShowtimeGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_LINE_SEC_HISTORY_STATS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecHistoryStats15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStatsDir_t));
         break;

      case DSL_FIO_PM_LINE_SEC_HISTORY_STATS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecHistoryStats1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStatsDir_t));
         break;

      case DSL_FIO_PM_LINE_SEC_THRESHOLDS_15MIN_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecThresholds15MinSet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecThreshold_t));
         break;

      case DSL_FIO_PM_LINE_SEC_THRESHOLDS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecThresholds15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecThreshold_t));
         break;

      case DSL_FIO_PM_LINE_SEC_THRESHOLDS_1DAY_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecThresholds1DaySet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecThreshold_t));
         break;

      case DSL_FIO_PM_LINE_SEC_THRESHOLDS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineSecThresholds1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineSecThreshold_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS*/

#ifdef INCLUDE_DSL_CPE_PM_HISTORY
      case DSL_FIO_PM_LINE_INIT_COUNTERS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitCounters15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitCounters_t));
         break;

      case DSL_FIO_PM_LINE_INIT_COUNTERS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitCounters1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
      case DSL_FIO_PM_LINE_INIT_COUNTERS_TOTAL_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitCountersTotalGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitCountersTotal_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
      case DSL_FIO_PM_LINE_INIT_COUNTERS_SHOWTIME_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitCountersShowtimeGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitCounters_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_LINE_INIT_HISTORY_STATS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitHistoryStats15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStats_t));
         break;

      case DSL_FIO_PM_LINE_INIT_HISTORY_STATS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitHistoryStats1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_HistoryStats_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
      case DSL_FIO_PM_LINE_INIT_THRESHOLDS_15MIN_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitThresholds15MinSet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitThreshold_t));
         break;

      case DSL_FIO_PM_LINE_INIT_THRESHOLDS_15MIN_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitThresholds15MinGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitThreshold_t));
         break;

      case DSL_FIO_PM_LINE_INIT_THRESHOLDS_1DAY_SET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_SET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitThresholds1DaySet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitThreshold_t));
         break;

      case DSL_FIO_PM_LINE_INIT_THRESHOLDS_1DAY_GET:
         nErrCode = DSL_DRV_IoctlHandleHelperCall(pContext, bIsInKernel,
            DSL_IOCTL_HELPER_GET,
            (DSL_DRV_IoctlHandlerHelperFunc_t)DSL_DRV_PM_LineInitThresholds1DayGet,
            (DSL_void_t*) nArg, sizeof(DSL_PM_LineInitThreshold_t));
         break;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS*/
#endif /* #if defined(INCLUDE_DSL_PM)*/
      default:
         nErrCode = DSL_ERR_NOT_SUPPORTED;
         break;
   }

   DSL_DRV_VFree(pIOCTL_arg);

   DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: OUT - DSL_DRV_IoctlHandle,"
      " return(from %d - %s) %d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nCommand, DSL_DBG_IoctlName(nCommand),
      nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_TestModeControlSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_TestModeControl_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_TestModeControlSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();


   if (pData->data.nTestMode >= DSL_TESTMODE_LAST || 
       pData->data.nTestMode < DSL_TESTMODE_DISABLE)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Wrong test mode (%d) specified!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pData->data.nTestMode));
	  return DSL_ERR_INVALID_PARAMETR;
   }

   /* Call device specific stuff*/
   nErrCode = DSL_DRV_DEV_TestModeControlSet(pContext,pData);

   if (pData->data.nTestMode == DSL_TESTMODE_DISABLE && nErrCode == DSL_SUCCESS)
   {
      /* Trigger restart sequence*/
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bAutobootRestart, DSL_TRUE);
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_TestModeControlSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_TestModeStatusGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_TestModeStatus_t *pData)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_TestModeStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_POINTER(pContext, pData);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ(pContext, nErrCode, nTestModeControl, pData->data.nTestMode);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_TestModeStatusGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_uint32_t DSL_DRV_LineStateUpdate(
   DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_LineStateValue_t nLineState = DSL_LINESTATE_UNKNOWN,
                        nPrevLineState = DSL_LINESTATE_UNKNOWN;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_LineStateUpdate"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Get Line State from the DSL CPE internal memory*/
   DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, nPrevLineState);
   if(nErrCode != DSL_SUCCESS)
   {
      return nErrCode;
   }

   nErrCode = DSL_DRV_DEV_LineStateGet(pContext, &nLineState);
   if (nErrCode == DSL_SUCCESS)
   {
      DSL_DRV_LineStateSet(pContext, nLineState);
   }
   else
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Could not get line state "
         "from the device!"DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_LineStateUpdate, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

DSL_DEV_Handle_t DSL_DRV_LowDeviceGet(DSL_devCtx_t *pDevCtx)
{
   return pDevCtx->lowHandle;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_FwDownload(
   DSL_Context_t *pContext,
   const DSL_char_t *pFw1,
   DSL_uint32_t nSize1,
   const DSL_char_t *pFw2,
   DSL_uint32_t nSize2,
   DSL_int32_t *pLoff,
   DSL_int32_t *pCurrentOff,
   DSL_boolean_t bLastChunk)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pContext->pDevCtx);
   DSL_CHECK_POINTER(pContext, pLoff);
   DSL_CHECK_POINTER(pContext, pCurrentOff);
   DSL_CHECK_ERR_CODE();

   DSL_DRV_LineStateSet(pContext, DSL_LINESTATE_IDLE_REQUEST);

   nErrCode = DSL_DRV_DEV_FwDownload(pContext,
         pFw1, nSize1, pFw2, nSize2, pLoff, pCurrentOff, bLastChunk);

   if (nErrCode != DSL_SUCCESS)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - firmware download has failed "
         "(on %d byte)"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), *pCurrentOff));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_MSG,
         (pContext, "DSL[%02d]: Firmware download finished "
         "(total %d bytes)"DSL_DRV_CRLF, DSL_DEV_NUM(pContext), *pCurrentOff));
   }

   return nErrCode;
}

DSL_Error_t DSL_DRV_LineStateSet(
   DSL_Context_t *pContext,
   DSL_LineStateValue_t nNewLineState)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_LineStateValue_t nOldLineState = DSL_LINESTATE_UNKNOWN;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, nOldLineState);

   if ((nNewLineState != nOldLineState) || (nNewLineState == DSL_LINESTATE_IDLE))
   {
      DSL_DEBUG(DSL_DBG_MSG, (pContext, "DSL[%02d]: Line state has changed: "
         "%08X -> %08X"DSL_DRV_CRLF, DSL_DEV_NUM(pContext),
         nOldLineState, nNewLineState));
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, nLineState, nNewLineState);

      nErrCode =  DSL_DRV_EventGenerate(
         pContext, 0, DSL_ACCESSDIR_NA, DSL_XTUDIR_NA,
         DSL_EVENT_S_LINE_STATE,
         (DSL_EventData_Union_t*)&nNewLineState,
         sizeof(DSL_LineStateData_t));

      if( nErrCode != DSL_SUCCESS )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - Event(%d) generate failed!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext), DSL_EVENT_S_LINE_STATE));
      }
   }

   return nErrCode;
}

DSL_Error_t DSL_DRV_ShowtimeStatusUpdate(
   DSL_Context_t *pContext,
   DSL_boolean_t bInit)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_LineStateValue_t nLineState = DSL_LINESTATE_UNKNOWN;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_ShowtimeStatusUpdate"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_CTX_READ_SCALAR(pContext, nErrCode, nLineState, nLineState);

   if(nLineState != DSL_LINESTATE_SHOWTIME_TC_SYNC &&
      nLineState != DSL_LINESTATE_SHOWTIME_NO_SYNC &&
      nLineState != DSL_LINESTATE_EXCHANGE)
   {
      DSL_DEBUG(DSL_DBG_ERR, (pContext, "DSL[%02d]: ERROR - This function is applicable "
         "while line is in SHOWTIME state only"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));
      nErrCode = DSL_ERR_ONLY_AVAILABLE_IN_SHOWTIME;
   }
   else
   {
      nErrCode = DSL_DRV_DEV_ShowtimeStatusUpdate(pContext, bInit);
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_ShowtimeStatusUpdate, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_CtxDataUpdate(
   DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint32_t tmp32 = 0;
   DSL_uint8_t i = 0;

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_CtxDataUpdate"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   for (i = 0 ; i < DSL_CHANNELS_PER_LINE; i++)
   {
      /* Update previous data rate values*/
      DSL_CTX_READ_SCALAR(pContext, nErrCode, nChannelActualDataRatePrevUs[i], tmp32);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, nChannelPreviousDataRateUs[i], tmp32);
      DSL_CTX_READ_SCALAR(pContext, nErrCode, nChannelActualDataRatePrevDs[i], tmp32);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, nChannelPreviousDataRateDs[i], tmp32);

      /* Reset actual data rate values*/
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, nChannelActualDataRateUs[i], 0);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, nChannelActualDataRateDs[i], 0);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, ActualInterleaveDelayUs[i], 0);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, ActualInterleaveDelayDs[i], 0);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, ActualImpulseNoiseProtectionUs[i], 0);
      DSL_CTX_WRITE_SCALAR(pContext, nErrCode, ActualImpulseNoiseProtectionDs[i], 0);
   }

   /* Clean current XTSE status*/
   DSL_DRV_MemSet(pContext->xtseCurr, 0, DSL_G997_NUM_XTSE_OCTETS);

#if (INCLUDE_DSL_CPE_API_VDSL_SUPPORT == 1)
   /* Clean current AUX line inventory information*/
   DSL_DRV_MemSet(&(pContext->auxInventoryFe), 0x0, sizeof(DSL_AuxLineInventoryData_t));   
#endif

   /* Clean current FE line inventory information*/
   DSL_DRV_MemSet(&(pContext->lineInventoryFe), 0x0, sizeof(DSL_G997_LineInventoryData_t));   
   /* Reset FE line inventory incomplete flag*/
   DSL_CTX_WRITE_SCALAR(pContext, nErrCode, bFeLineInventoryIncomplete, DSL_FALSE);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_CtxDataUpdate, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/* Events stuff */

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_EventQueue(
   DSL_Context_t *pContext,
   DSL_IN DSL_EventStatusData_t *pEvent,
   DSL_IN DSL_uint32_t nDataSize)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint8_t *buf;
   DSL_EventStatusData_t *pEvBuf;
   DSL_OpenContext_t *pCurr;

   DSL_DEBUG(DSL_DBG_MSG,
      (DSL_NULL, "DSL[%02d]: IN - DSL_DRV_EventQueue"DSL_DRV_CRLF,
      pOpenContextList->pDevCtx->nNum));
   
   if (pEvent == DSL_NULL)
   {
      return DSL_ERR_INVALID_PARAMETR;
   }

   for (pCurr = pOpenContextList; pCurr != DSL_NULL; pCurr = pCurr->pNext)
   {
      /* Queue event only for a specified device*/
      if (pContext->pDevCtx->nNum != pCurr->pDevCtx->nNum)
      {
         continue;
      }

      if(DSL_DRV_MUTEX_LOCK(pCurr->eventMutex))
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (DSL_NULL, "DSL[%02d]: ERROR - Couldn't lock event mutex!"DSL_DRV_CRLF,
            pOpenContextList->pDevCtx->nNum));
      }
      else
      {
         if ((pCurr->eventFifo != DSL_NULL) &&
             (pCurr->eventFifoBuf != DSL_NULL) &&
             (pCurr->bEventActivation == DSL_TRUE) &&
             (pCurr->nEventMask & DSL_EVENT2MASK(pEvent->nEventType)) == 0)
         {
            if ( DSL_Fifo_isFull( pCurr->eventFifo ) == DSL_TRUE )
            {
               DSL_DEBUG( DSL_DBG_WRN,
                  (DSL_NULL, "DSL[%02d]: WARNING - The fifo is full!"DSL_DRV_CRLF,
                  pOpenContextList->pDevCtx->nNum));
            }
            else
            {
               if ((buf = DSL_Fifo_writeElement(pCurr->eventFifo)) == DSL_NULL)
               {
                  DSL_DEBUG( DSL_DBG_ERR,
                     (DSL_NULL, "DSL[%02d]: ERROR - Event could not be added!"DSL_DRV_CRLF,
                     pOpenContextList->pDevCtx->nNum));
               }
               else
               {
                  /* copy an element */
                  memcpy(buf, pEvent, sizeof(DSL_EventStatusData_t));

                  if (pEvent->pData != DSL_NULL)
                  {
                     pEvBuf = (DSL_EventStatusData_t *)buf;
                     pEvBuf->pData = DSL_DRV_VMalloc(nDataSize);
                     if (pEvBuf->pData != DSL_NULL)
                     {
                        /* copy an element data */
                        memcpy(pEvBuf->pData, pEvent->pData, nDataSize);
                     }
                     else
                     {
                        DSL_DEBUG( DSL_DBG_WRN,
                           (DSL_NULL, "DSL[%02d]: WRN - Could not allocate memory for an "
                           "event data, it will be lost!"DSL_DRV_CRLF,
                           pOpenContextList->pDevCtx->nNum));
                     }
                  }

                  /* wake up */
                  DSL_DRV_WAKEUP_WAKELIST(pCurr->eventWaitQueue);
               }
            }
         }

         DSL_DRV_MUTEX_UNLOCK(pCurr->eventMutex);
      }
   }

   /* free event data */
   if (pEvent->pData != DSL_NULL)
   {
      DSL_DRV_VFree(pEvent->pData);
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (DSL_NULL, "DSL[%02d]: OUT - DSL_DRV_EventQueue, retCode=%d"DSL_DRV_CRLF,
      pOpenContextList->pDevCtx->nNum, nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern.h'
*/
DSL_Error_t DSL_DRV_EventUnqueue(
   DSL_IN DSL_OpenContext_t *pOpenContext,
   DSL_IN DSL_EventStatusData_t *pEvent)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_uint8_t *buf;

   if (pOpenContext == DSL_NULL || pEvent == DSL_NULL)
   {
      return DSL_ERR_INVALID_PARAMETR;
   }

   if (pOpenContext->eventFifo == DSL_NULL ||
      pOpenContext->eventFifoBuf == DSL_NULL)
   {
      DSL_DEBUG( DSL_DBG_WRN,
         (DSL_NULL, "DSL[%02d]: WRN - Call to event routine before it "
         "is configured!"DSL_DRV_CRLF, pOpenContext->pDevCtx->nNum));
      return DSL_ERROR;
   }

   if(DSL_DRV_MUTEX_LOCK(pOpenContext->eventMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (DSL_NULL, "DSL[%02d]: ERROR - Couldn't lock event mutex!"DSL_DRV_CRLF,
         pOpenContext->pDevCtx->nNum));
      return DSL_ERROR;
   }

   if ( DSL_Fifo_isEmpty( pOpenContext->eventFifo ) == DSL_TRUE )
   {
      DSL_DEBUG( DSL_DBG_WRN,
         (DSL_NULL, "DSL[%02d]: WARNING - The fifo is empty"DSL_DRV_CRLF,
         pOpenContext->pDevCtx->nNum));
      /* FIXME: in this place a warning code should be returned,
         it needs some fixes in the upper code */
      nErrCode = DSL_SUCCESS;
   }
   else
   {
      if ((buf = DSL_Fifo_readElement(pOpenContext->eventFifo)) == DSL_NULL)
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (DSL_NULL, "DSL[%02d]: ERROR - Event could not be read!"DSL_DRV_CRLF,
            pOpenContext->pDevCtx->nNum));
         nErrCode = DSL_ERROR;
      }
      else
      {
         /* copy an element */
         memcpy(pEvent, buf, sizeof(DSL_EventStatusData_t));
      }
   }

   DSL_DRV_MUTEX_UNLOCK(pOpenContext->eventMutex);

   DSL_DEBUG(DSL_DBG_MSG,
      (DSL_NULL, "DSL[%02d]: OUT - DSL_DRV_EventUnqueue, retCode=%d"DSL_DRV_CRLF,
      pOpenContext->pDevCtx->nNum, nErrCode));

   return nErrCode;
}

/* Helpers stuff */

/*
   This function converts a given value in unsigned 16-bit value in 2'complement
   format to its according signed 16-bit format.
   The type of 2'complement can be specified, for example:
   nsVal16 = 0x200, nBits = 10 ==> nuVal16 = -512

   \param nsVal16
      Specifies the unsigned 16-bit value that shall be converted, [I]
   \param nBits
      Specifies how many bits are used for 2'complement conversion.
      The bit that includes the sign is included, [I]
      \note It makes only sense to use this function in a range of
            2 <= nBits <= 15
            In case of nBits = 16 a simple cast will do the right thing.
   \param nuVal16
      Returns the converted 16-bit int value, [O]

   \return
   Return values are defined within the DSL_Error_t definition
   - DSL_SUCCESS in case of success
   - DSL_ERROR if operation failed
*/
DSL_Error_t DSL_DRV_TwosComplement16_HexToInt(
   DSL_Context_t *pContext,
   DSL_uint16_t nuVal16,
   DSL_char_t nBits,
   DSL_int16_t *nsVal16)
{
   DSL_uint16_t nFullScale = 0, nFullScaleMask = 0;

   *nsVal16 = 0;

   if ((nBits < 2) || (nBits > 16))
   {
      DSL_DEBUG(DSL_DBG_ERR, (pContext, "DSL[%02d]: Type of 2'complement (%d)"
         " conversion out of valid range (2...16)"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), nBits));

      return DSL_ERROR;
   }

   nFullScale = (DSL_uint16_t) (1 << (nBits - 1));
   nFullScaleMask = nFullScale - 1;

   /* If msb is '1' it is a negative value */
   if (nuVal16 > (nFullScale - 1))
   {
      *nsVal16 = (DSL_int16_t) (nFullScaleMask & nuVal16);
      *nsVal16 -= ((DSL_int16_t) nFullScale);
   }
   else
   {
      *nsVal16 = (DSL_int16_t) nuVal16;
   }

   return DSL_SUCCESS;
}

#ifndef DRV_DSL_CPE_FORCE_MACROS
DSL_Error_t
_DSL_CHECK_CTX_POINTER(
   DSL_Context_t *pContext)
{
   if (pContext == DSL_NULL)
   {
      DSL_DEBUG_SET_ERROR(DSL_ERR_POINTER);
      DSL_DEBUG_HDR(DSL_DBG_ERR, (pContext, "Invalid context pointer!"));
      return DSL_ERR_POINTER;
   }

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CHECK_POINTER(
   DSL_Context_t *pContext,
   DSL_void_t *pPtr)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   if (pPtr == DSL_NULL)
   {
      DSL_DEBUG_SET_ERROR(DSL_ERR_POINTER);
      DSL_DEBUG_HDR(DSL_DBG_ERR, (pContext, "Invalid data pointer!"));
      return DSL_ERR_POINTER;
   }

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CHECK_MODEM_IS_READY(
   DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pContext->pDevCtx);
   DSL_CHECK_ERR_CODE();

   if (DSL_DRV_DEV_ModemIsReady(pContext) == DSL_FALSE)
   {
      pContext->nErrNo = DSL_ERR_MODEM_NOT_READY;
      DSL_DEBUG_HDR(DSL_DBG_ERR, (pContext, "Modem is not ready!")); \
      return DSL_ERR_MODEM_NOT_READY;
   }

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CHECK_CHANNEL_RANGE(
   DSL_Context_t *pContext,
   DSL_uint16_t nChannel)
{
  if (nChannel >= DSL_CHANNELS_PER_LINE)
  {
     DSL_DEBUG_SET_ERROR(DSL_ERR_CHANNEL_RANGE);
     DSL_DEBUG(DSL_DBG_ERR,
        (pContext, "DSL: invalid bearer channel number %d! (only bearer "
         "channels in the range of 0..%d are supported)" DSL_DRV_CRLF,
         nChannel, (DSL_CHANNELS_PER_LINE - 1)));
      return DSL_ERR_CHANNEL_RANGE;
   }

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CHECK_ATU_DIRECTION(
   DSL_Context_t *pContext,
   DSL_XTUDir_t nDirection)
{
   if ((nDirection != DSL_NEAR_END) && (nDirection != DSL_FAR_END))
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL: invalid ATU direction!" DSL_DRV_CRLF));
      return DSL_ERR_DIRECTION;
   }

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CHECK_DIRECTION(
   DSL_Context_t *pContext,
   DSL_AccessDir_t nDirection)
{
   if ((nDirection != DSL_DOWNSTREAM) && (nDirection != DSL_UPSTREAM))
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL: invalid direction!" DSL_DRV_CRLF));
      return DSL_ERR_DIRECTION;
   }

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CTX_ASSIGN_ANY(
   DSL_Context_t *pContext,
   DSL_void_t *pFrom,
   DSL_void_t *pTo,
   DSL_uint32_t nSize)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pFrom);
   DSL_CHECK_POINTER(pContext, pTo);
   if (nSize == 0)
   {
      DSL_DEBUG( DSL_DBG_ERR, (pContext, "_DSL_CTX_ASSIGN_ANY: "
         "nSize = 0!!!"DSL_DRV_CRLF));
      nErrCode |= DSL_ERR_INVALID_PARAMETR;
   }
   DSL_CHECK_ERR_CODE();

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR, (pContext, "Couldn't lock data mutex "
         "(pContext=%p, pFrom=%p, pTo=%p"DSL_DRV_CRLF, pContext, pFrom, pTo));
      return DSL_ERR_SEMAPHORE_GET;
   }

   memcpy(pTo, pFrom, nSize);

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CTX_ASSIGN_8(
   DSL_Context_t *pContext,
   DSL_uint8_t nFrom,
   DSL_void_t *pTo)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pTo);
   DSL_CHECK_ERR_CODE();

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR, (pContext, "Couldn't lock data mutex "
         "(pContext=%p, nFrom=%02x, pTo=%p"DSL_DRV_CRLF, pContext, nFrom, pTo));
      return DSL_ERR_SEMAPHORE_GET;
   }

   *((DSL_uint8_t*)pTo) = nFrom;

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CTX_ASSIGN_16(
   DSL_Context_t *pContext,
   DSL_uint16_t nFrom,
   DSL_void_t *pTo)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pTo);
   DSL_CHECK_ERR_CODE();

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR, (pContext, "Couldn't lock data mutex "
         "(pContext=%p, nFrom=%04x, pTo=%p"DSL_DRV_CRLF, pContext, nFrom, pTo));
      return DSL_ERR_SEMAPHORE_GET;
   }

   *((DSL_uint16_t*)pTo) = nFrom;

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   return DSL_SUCCESS;
}

DSL_Error_t
_DSL_CTX_ASSIGN_32(
   DSL_Context_t *pContext,
   DSL_uint32_t nFrom,
   DSL_void_t *pTo)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pTo);
   DSL_CHECK_ERR_CODE();

   if(DSL_DRV_MUTEX_LOCK(pContext->dataMutex))
   {
      DSL_DEBUG( DSL_DBG_ERR, (pContext, "Couldn't lock data mutex "
         "(pContext=%p, nFrom=%08x, pTo=%p"DSL_DRV_CRLF, pContext, nFrom, pTo));
      return DSL_ERR_SEMAPHORE_GET;
   }

   *((DSL_uint32_t*)pTo) = nFrom;

   DSL_DRV_MUTEX_UNLOCK(pContext->dataMutex);

   return DSL_SUCCESS;
}

#endif

#ifdef __cplusplus
}
#endif
